import numpy as np
import pytest
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_random_state

import fri
from fri import FRI, genLupiData


@pytest.fixture(scope="function")
def randomstate():
    return check_random_state(1337)


@pytest.mark.parametrize("problem", fri.LUPI_MODELS)
def test_error_class(problem):
    X, X_priv, y = genLupiData(problem, n_samples=500, n_strel=3)

    f = FRI(problem)

    combined = np.hstack([X, X_priv])

    with pytest.raises(ValueError):
        f.fit(combined, y, lupi_features=0)
    d = combined.shape[1]
    with pytest.raises(ValueError):
        f.fit(combined, y, lupi_features=d)
    with pytest.raises(ValueError):
        f.fit(combined, y)


@pytest.mark.parametrize("n_weak", [0, 2])
@pytest.mark.parametrize("problem", fri.LUPI_MODELS)
def test_lupi_model_correctness(problem, n_weak, randomstate):
    n_samples = 100

    model = FRI(
        problem,
        random_state=randomstate,
        verbose=0,
        # slack_regularization=0.1,
        # slack_loss=0.1,
        n_param_search=50,
        n_probe_features=70,
        n_jobs=1,
    )
    print(model)

    data = genLupiData(
        problem,
        n_strel=1,
        n_weakrel=n_weak,
        n_samples=n_samples,
        n_irrel=2,
        # label_noise=0.1,
        n_repeated=0,
        random_state=randomstate,
    )

    X, X_priv, y = data

    n_priv_features = X_priv.shape[1]
    X = StandardScaler().fit(X).transform(X)
    X_priv = StandardScaler().fit(X_priv).transform(X_priv)
    combined = np.hstack([X, X_priv])

    model.fit(combined, y, lupi_features=n_priv_features)

    # Check the interval output
    interval = model.interval_
    print(interval)
    print(model.print_interval_with_class())
    assert len(model.allrel_prediction_) == X.shape[1] + X_priv.shape[1]
    assert len(interval) == X.shape[1] + X_priv.shape[1]



import cvxpy as cvx
import numpy as np
from sklearn import preprocessing
from sklearn.metrics import fbeta_score, classification_report
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import check_X_y
from sklearn.utils.multiclass import unique_labels

from .base_initmodel import InitModel
from .base_lupi import LUPI_Relevance_CVXProblem, split_dataset
from .base_type import ProblemType
from .classification import Classification_Relevance_Bound


class LUPI_Classification(ProblemType):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._lupi_features = None

    @property
    def lupi_features(self):
        return self._lupi_features

    @classmethod
    def parameters(cls):
        return ["C", "scaling_lupi_w", "scaling_lupi_loss"]

    @property
    def get_initmodel_template(cls):
        return LUPI_Classification_SVM

    @property
    def get_cvxproblem_template(cls):
        return LUPI_Classification_Relevance_Bound

    def relax_factors(cls):
        return ["loss_slack", "w_l1_slack"]

    def preprocessing(self, data, lupi_features=None):
        X, y = data
        d = X.shape[1]

        if lupi_features is None:
            raise ValueError("Argument 'lupi_features' missing in fit() call.")
        if not isinstance(lupi_features, int):
            raise ValueError("Argument 'lupi_features' is not type int.")
        if not 0 < lupi_features < d:
            raise ValueError(
                "Argument 'lupi_features' looks wrong. We need at least 1 priviliged feature (>0) or at least one normal feature."
            )

        self._lupi_features = lupi_features

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit
        classes_ = unique_labels(y)

        if len(classes_) > 2:
            raise ValueError("Only binary class data supported")

        # Negative class is set to -1 for decision surface
        y = preprocessing.LabelEncoder().fit_transform(y)
        y[y == 0] = -1

        return X, y


class LUPI_Classification_SVM(InitModel):
    @classmethod
    def hyperparameter(cls):
        return ["C", "scaling_lupi_w", "scaling_lupi_loss"]

    def fit(self, X_combined, y, lupi_features=None):
        """

        Parameters
        ----------
        lupi_features : int
            Number of features in dataset which are considered privileged information (PI).
            PI features are expected to be the last features in the dataset.

        """
        if lupi_features is None:
            raise ValueError("No lupi_features argument given.")
        self.lupi_features = lupi_features
        X, X_priv = split_dataset(X_combined, lupi_features)
        (n, d) = X.shape

        # Get parameters from CV model without any feature contstraints
        C = self.hyperparam["C"]
        scaling_lupi_w = self.hyperparam["scaling_lupi_w"]
        scaling_lupi_loss = self.hyperparam["scaling_lupi_loss"]

        # Initalize Variables in cvxpy
        w = cvx.Variable(shape=(d), name="w")
        w_priv = cvx.Variable(lupi_features, name="w_priv")
        b = cvx.Variable(name="bias")
        b_priv = cvx.Variable(name="bias_priv")

        # Define functions for better readability
        function = X * w + b
        priv_function = X_priv * w_priv + b_priv
        slack = cvx.Variable(shape=(n))

        # Combined loss of lupi function and normal slacks, scaled by two constants
        loss = scaling_lupi_loss * cvx.sum(priv_function) + cvx.sum(slack)

        # L1 norm regularization of both functions with 1 scaling constant
        w_l1 = cvx.norm(w, 1)
        w_priv_l1 = cvx.norm(w_priv, 1)
        weight_regularization = 0.5 * (w_l1 + scaling_lupi_w * w_priv_l1)

        constraints = [
            cvx.multiply(y.T, function) >= 1 - cvx.multiply(y.T, priv_function) - slack,
            priv_function >= 0,
            slack >= 0,
        ]
        objective = cvx.Minimize(C * loss + weight_regularization)

        # Solve problem.
        solver_params = self.solver_params
        problem = cvx.Problem(objective, constraints)
        problem.solve(**solver_params)

        w = w.value
        w_priv = w_priv.value
        b = b.value
        b_priv = b_priv.value
        self.model_state = {
            "w": w,
            "w_priv": w_priv,
            "b": b,
            "b_priv": b_priv,
            "lupi_features": lupi_features,  # Number of lupi features in the dataset TODO: Move this somewhere else
        }

        loss = loss.value
        w_l1 = w_l1.value
        w_priv_l1 = w_priv_l1.value
        self.constraints = {"loss": loss, "w_l1": w_l1, "w_priv_l1": w_priv_l1}
        return self

    def predict(self, X):
        X, X_priv = split_dataset(X, self.lupi_features)
        w = self.model_state["w"]
        b = self.model_state["b"]

        # Simple hyperplane classification rule
        f = np.dot(X, w) + b
        y = f >= 0
        y = y.astype(int)

        # Format binary as signed unit vector
        y[y == 0] = -1

        return y

    def score(self, X, y, **kwargs):
        prediction = self.predict(X)

        # Negative class is set to -1 for decision surface
        y = LabelEncoder().fit_transform(y)
        y[y == 0] = -1

        # Using weighted f1 score to have a stable score for imbalanced datasets
        score = fbeta_score(y, prediction, beta=1, average="weighted")
        if "verbose" in kwargs:
            return classification_report(y, prediction)
        return score


class LUPI_Classification_Relevance_Bound(
    LUPI_Relevance_CVXProblem, Classification_Relevance_Bound
):
    def _init_objective_UB_LUPI(self, sign=None, **kwargs):
        self.add_constraint(
            self.feature_relevance <= sign * self.w_priv[self.lupi_index]
        )
        self._objective = cvx.Maximize(self.feature_relevance)

    def _init_objective_LB_LUPI(self, **kwargs):
        self.add_constraint(
            cvx.abs(self.w_priv[self.lupi_index]) <= self.feature_relevance
        )
        self._objective = cvx.Minimize(self.feature_relevance)

    def _init_constraints(self, parameters, init_model_constraints):
        # Upper constraints from best initial model
        l1_w = init_model_constraints["w_l1"]
        l1_priv_w = init_model_constraints["w_priv_l1"]
        init_loss = init_model_constraints["loss"]

        # New Variables
        w = cvx.Variable(shape=(self.d), name="w")
        w_priv = cvx.Variable(shape=(self.d_priv), name="w_priv")
        b = cvx.Variable(name="b")
        b_priv = cvx.Variable(name="b_priv")
        slack = cvx.Variable(shape=(self.n))

        # New Constraints
        function = cvx.multiply(self.y.T, self.X * w + b)
        priv_function = self.X_priv * w_priv + b_priv
        loss = cvx.sum(priv_function) + cvx.sum(slack)

        weight_norm = cvx.norm(w, 1)
        weight_norm_priv = cvx.norm(w_priv, 1)

        self.add_constraint(
            function >= 1 - cvx.multiply(self.y.T, priv_function) - slack
        )
        self.add_constraint(priv_function >= 0)
        self.add_constraint(loss <= init_loss)
        self.add_constraint(weight_norm + weight_norm_priv <= l1_w + l1_priv_w)
        self.add_constraint(slack >= 0)

        # Save values for object use later
        self.w = w
        self.w_priv = w_priv
        self.feature_relevance = cvx.Variable(nonneg=True, name="Feature Relevance")



"""
    In this class we use hyperparameter search to find parameters needed in our model.
    Depending on the input model we sample parameters from a random distribution.
    The sampling rate can be increased.
    The model with the best internally defined accuracy is picked.
    To increase robustness we use cross validation.
"""
import warnings

from sklearn.exceptions import FitFailedWarning

warnings.filterwarnings(action="ignore", category=FitFailedWarning)

from pprint import pprint
from typing import Tuple

import numpy as np
from sklearn.model_selection import RandomizedSearchCV

from fri.model.base_initmodel import InitModel


def find_best_model(
    model_template: InitModel,
    hyperparameters: dict,
    data: Tuple[np.ndarray, np.ndarray],
    random_state: np.random.RandomState,
    n_iter: int,
    n_jobs: int,
    verbose: int = 0,
    lupi_features=None,
    kwargs: dict = None,
) -> Tuple[InitModel, float]:
    """
    Search function which wraps `sklearns`  `RandomizedSearchCV` function.
    We use distributions and parameters defined in the `model_template`.

    Parameters
    ----------
    model_template : InitModel
        A model template which is used to fit data.
    hyperparameters : dict
        Dictionary of hyperparameters.
    data : tuple
        Tuple of data (X,y)
    random_state : RandomState
        numpy RandomState object
    n_iter : int
        Amount of search samples.
    n_jobs : int
        Allows multiprocessing with `n_jobs` threads.
    verbose : int
        Allows verbose output when `verbose>0`.
    lupi_features : int
        Amount of lupi_features
    kwargs : dict
        Placeholder, dict to pass into fit functions.
    """
    model = model_template()

    scorer, metric = model.make_scorer()
    if scorer is None:
        refit = True
    else:
        refit = metric

    searcher = RandomizedSearchCV(
        model,
        hyperparameters,
        scoring=scorer,
        random_state=random_state,
        refit=refit,
        cv=3,
        n_iter=n_iter,
        n_jobs=n_jobs,
        error_score=np.nan,
        verbose=verbose,
    )

    X, y = data
    # Ignore warnings for extremely bad model_state (when precision=0)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        searcher.fit(X, y, lupi_features=lupi_features)

    best_model: InitModel = searcher.best_estimator_
    best_score = best_model.score(X, y)
    if verbose > 0:
        print("*" * 20, "Best found baseline model", "*" * 20)
        pprint(best_model)
        print("score: ", best_score)
        for k, v in best_model.constraints.items():
            pprint((f"{k}: {v}"))
        for k, v in best_model.model_state.items():
            if hasattr(v, "shape"):
                pprint((f"{k}: shape {v.shape}"))
            else:
                if "slack" in k:
                    continue
                pprint((f"{k}: {v}"))
        print("*" * 30)
    return best_model, best_score



from abc import ABC, abstractmethod

import cvxpy as cvx
import numpy as np
from cvxpy import SolverError


class Relevance_CVXProblem(ABC):
    def __str__(self) -> str:
        if self.isLowerBound:
            lower = "Lower"
        else:
            lower = "Upper"
        name = f"{lower}_{self.current_feature}_{self.__class__.__name__}"
        state = ""
        for s in self.init_hyperparameters.items():
            state += f"{s[0]}:{s[1]}, "
        for s in self.init_model_constraints.items():
            state += f"{s[0]}:{s[1]}, "
        state = "(" + state[:-2] + ")"
        if self.isProbe:
            prefix = f"Probe_{self.probeID}"
        else:
            prefix = ""
        return prefix + name + state

    def __init__(
        self,
        current_feature: int,
        data: tuple,
        hyperparameters,
        best_model_constraints,
        preset_model=None,
        best_model_state=None,
        probeID=-1,
        **kwargs,
    ) -> None:
        self._probeID = probeID
        self._feature_relevance = None
        self.isLowerBound = None

        # General data
        self.current_feature = current_feature
        self.preset_model = preset_model
        self.best_model_state = best_model_state

        self.preprocessing_data(data, best_model_state)

        # Initialize constraints
        self._constraints = []
        self._objective = None
        self.w = None
        self._init_constraints(hyperparameters, best_model_constraints)

        if self.preset_model is not None:
            self._add_preset_constraints(self.preset_model, best_model_constraints)

        self.init_hyperparameters = hyperparameters
        self.init_model_constraints = best_model_constraints

    def preprocessing_data(self, data, best_model_state):
        X, y = data
        self.n = X.shape[0]
        self.d = X.shape[1]
        self.X = X
        self.y = np.array(y)

    @property
    def constraints(self):
        return self._constraints

    def add_constraint(self, new):
        self._constraints.append(new)

    @property
    def objective(self):
        return self._objective

    @property
    def solved_relevance(self):
        if self.is_solved:
            return self.objective.value
        else:
            raise Exception("Problem not solved. No feature relevance computed.")

    @property
    def probeID(self):
        return self._probeID

    @property
    def isProbe(self):
        return self.probeID >= 0

    @abstractmethod
    def _init_constraints(self, parameters, init_model_constraints):
        pass

    @abstractmethod
    def init_objective_UB(self, **kwargs):
        pass

    @abstractmethod
    def init_objective_LB(self, **kwargs):
        pass

    @property
    def cvx_problem(self):
        return self._cvx_problem

    @property
    def is_solved(self):
        if self._solver_status in self.accepted_status:
            try:
                val = self.objective.value
            except ValueError:
                return False

            return True
        else:
            return False

    @property
    def accepted_status(self):
        return ["optimal", "optimal_inaccurate"]

    def solve(self) -> object:
        # We init cvx problem here because pickling LP solver objects is problematic
        # by deferring it to here, worker threads do the problem building themselves and we spare the serialization
        self._cvx_problem = cvx.Problem(
            objective=self.objective, constraints=self.constraints
        )
        try:
            # print("Solve", self)
            self._cvx_problem.solve(**self.solver_kwargs)
        except SolverError:
            # We ignore Solver Errors, which are common with our framework:
            # We solve multiple problems per bound and choose a feasible solution later (see '_create_interval')
            pass

        self._solver_status = self._cvx_problem.status
        # self._cvx_problem = None
        return self

    def _retrieve_result(self):
        return self.current_feature, self.objective

    @property
    def solver_kwargs(self):
        return {"verbose": False, "solver": "ECOS"}

    def _add_preset_constraints(self, preset_model: dict, best_model_constraints):

        for feature, current_preset in preset_model.items():
            # Skip current feature
            if feature == self.current_feature:
                continue

            # Skip unset values
            if all(np.isnan(current_preset)):
                continue

            # a weight bigger than the optimal model L1 makes no sense
            assert abs(current_preset[0]) <= best_model_constraints["w_l1"]
            assert abs(current_preset[1]) <= best_model_constraints["w_l1"]

            # We add a pair of constraints depending on sign of known coefficient
            # this makes it possible to solve this as a convex problem
            if current_preset[0] >= 0:
                self.add_constraint(self.w[feature] >= current_preset[0])
                self.add_constraint(self.w[feature] <= current_preset[1])
            else:
                self.add_constraint(self.w[feature] <= current_preset[0])
                self.add_constraint(self.w[feature] >= current_preset[1])

    @classmethod
    def generate_lower_bound_problem(
        cls,
        best_hyperparameters,
        init_constraints,
        best_model_state,
        data,
        di,
        preset_model,
        probeID=-1,
    ):
        problem = cls(
            di,
            data,
            best_hyperparameters,
            init_constraints,
            preset_model=preset_model,
            best_model_state=best_model_state,
            probeID=probeID,
        )
        problem.init_objective_LB()
        problem.isLowerBound = True
        yield problem

    @classmethod
    def generate_upper_bound_problem(
        cls,
        best_hyperparameters,
        init_constraints,
        best_model_state,
        data,
        di,
        preset_model,
        probeID=-1,
    ):
        for sign in [-1, 1]:
            problem = cls(
                di,
                data,
                best_hyperparameters,
                init_constraints,
                preset_model=preset_model,
                best_model_state=best_model_state,
                probeID=probeID,
            )
            problem.init_objective_UB(sign=sign)
            problem.isLowerBound = False
            yield problem

    @classmethod
    def aggregate_min_candidates(cls, min_problems_candidates):
        vals = [candidate.solved_relevance for candidate in min_problems_candidates]
        min_value = min(vals)
        return min_value

    @classmethod
    def aggregate_max_candidates(cls, max_problems_candidates):
        vals = [candidate.solved_relevance for candidate in max_problems_candidates]
        max_value = max(vals)
        return max_value



import pytest
import numpy as np

from fri.model.ordinal_regression import ordinal_scores as score


@pytest.fixture()
def data():
    y = np.array([0, 1, 2, 3])
    return y


@pytest.fixture()
def imb_data():
    y = np.array([0, 1, 2, 2, 2, 2])
    return y


def reverse_label(y):
    y = np.copy(y)
    return np.flip(y[:], axis=0)


def swap_first_last(y):
    y = np.copy(y)
    y[[0, -1]] = y[[-1, 0]]
    return y


@pytest.mark.parametrize("error", ["mze", "mae", "mmae"])
def test_ordinal_score(error, data):
    score_perfect = score(data, data, error_type=error)
    score_mixed = score(data, swap_first_last(data), error_type=error)
    score_worst = score(data, reverse_label(data), error_type=error)

    assert score_perfect > score_mixed
    assert score_mixed > score_worst

    assert score_perfect == 1


def test_score_imbalanced(data, imb_data):
    score_mae = score(data, swap_first_last(data), error_type="mae")
    score_mmae = score(data, swap_first_last(data), error_type="mmae")

    assert score_mae == score_mmae

    score_mae = score(imb_data, swap_first_last(imb_data), error_type="mae")
    score_mmae = score(imb_data, swap_first_last(imb_data), error_type="mmae")

    assert score_mae != score_mmae



from abc import ABC, abstractmethod

import scipy.stats


class ProblemType(ABC):
    def __init__(self, **kwargs):

        self.chosen_parameters_ = {}

        for p in self.parameters():
            if p in kwargs:
                if kwargs[p] is not None:
                    self.chosen_parameters_[p] = kwargs[p]

        self.relax_factors_ = {}
        for p in self.relax_factors():
            if p in kwargs:
                if kwargs[p] is not None:
                    self.relax_factors_[p] = kwargs[p]

    @classmethod
    @abstractmethod
    def parameters(cls):
        raise NotImplementedError

    def get_chosen_parameter(self, p):
        try:
            return [
                self.chosen_parameters_[p]
            ]  # We return list for param search function
        except:
            # # TODO: rewrite the parameter logic
            # # TODO: move this to subclass
            if p == "scaling_lupi_w":
                # return [0.1, 1, 10, 100, 1000]
                return scipy.stats.reciprocal(a=1e-15, b=1e10)
            # if p == "scaling_lupi_loss":
            #    # value 0>p<1 causes standard svm solution
            #    # p>1 encourages usage of lupi function
            #    return scipy.stats.reciprocal(a=1e-15, b=1e15)
            if p == "C":
                return scipy.stats.reciprocal(a=1e-5, b=1e5)
            if p == "epsilon":
                return [0, 0.001, 0.01, 0.1, 1, 10, 100]
            else:
                return scipy.stats.reciprocal(a=1e-10, b=1e10)

    def get_all_parameters(self):
        return {p: self.get_chosen_parameter(p) for p in self.parameters()}

    @classmethod
    @abstractmethod
    def relax_factors(cls):
        raise NotImplementedError

    def get_chosen_relax_factors(self, p):
        try:
            factor = self.relax_factors_[p]
        except KeyError:
            try:
                factor = self.relax_factors_[p + "_slack"]
            except KeyError:
                factor = 0.1
        if factor < 0:
            raise ValueError("Slack Factor multiplier is positive!")
        return factor

    def get_all_relax_factors(self):
        return {p: self.get_chosen_relax_factors(p) for p in self.relax_factors()}

    @property
    @abstractmethod
    def get_initmodel_template(self):
        pass

    @property
    @abstractmethod
    def get_cvxproblem_template(self):
        pass

    @abstractmethod
    def preprocessing(self, data, lupi_features=None):
        return data

    def postprocessing(self, bounds):
        return bounds

    def get_relaxed_constraints(self, constraints):
        return {c: self.relax_constraint(c, v) for c, v in constraints.items()}

    def relax_constraint(self, key, value):
        return value * (1 + self.get_chosen_relax_factors(key))



from itertools import product

import cvxpy as cvx
import numpy as np
from sklearn.metrics import r2_score
from sklearn.metrics.regression import _check_reg_targets
from sklearn.utils import check_X_y

from fri.model.base_lupi import (
    LUPI_Relevance_CVXProblem,
    split_dataset,
    is_lupi_feature,
)
from fri.model.regression import Regression_Relevance_Bound
from .base_initmodel import LUPI_InitModel
from .base_type import ProblemType


class LUPI_Regression(ProblemType):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._lupi_features = None

    @property
    def lupi_features(self):
        return self._lupi_features

    @classmethod
    def parameters(cls):
        return ["C", "epsilon", "scaling_lupi_w", "scaling_lupi_loss"]

    @property
    def get_initmodel_template(cls):
        return LUPI_Regression_SVM

    @property
    def get_cvxproblem_template(cls):
        return LUPI_Regression_Relevance_Bound

    def relax_factors(cls):
        return ["loss_slack", "w_l1_slack"]

    def preprocessing(self, data, lupi_features=None):
        X, y = data
        d = X.shape[1]

        if lupi_features is None:
            raise ValueError("Argument 'lupi_features' missing in fit() call.")
        if not isinstance(lupi_features, int):
            raise ValueError("Argument 'lupi_features' is not type int.")
        if not 0 < lupi_features < d:
            raise ValueError(
                "Argument 'lupi_features' looks wrong. We need at least 1 priviliged feature (>0) or at least one normal feature."
            )

        self._lupi_features = lupi_features

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        return X, y


class LUPI_Regression_SVM(LUPI_InitModel):
    @classmethod
    def hyperparameter(cls):
        return ["C", "epsilon", "scaling_lupi_w", "scaling_lupi_loss"]

    def fit(self, X_combined, y, lupi_features=None):
        """

        Parameters
        ----------
        lupi_features : int
            Number of features in dataset which are considered privileged information (PI).
            PI features are expected to be the last features in the dataset.

        """
        if lupi_features is None:
            raise ValueError("No lupi_features argument given.")
        self.lupi_features = lupi_features
        X, X_priv = split_dataset(X_combined, lupi_features)
        (n, d) = X.shape

        # Get parameters from CV model without any feature contstraints
        C = self.hyperparam["C"]
        epsilon = self.hyperparam["epsilon"]
        scaling_lupi_w = self.hyperparam["scaling_lupi_w"]
        scaling_lupi_loss = self.hyperparam["scaling_lupi_loss"]

        # Initalize Variables in cvxpy
        w = cvx.Variable(shape=(d), name="w")
        b = cvx.Variable(name="bias")
        w_priv_pos = cvx.Variable(lupi_features, name="w_priv_pos")
        b_priv_pos = cvx.Variable(name="bias_priv_pos")
        w_priv_neg = cvx.Variable(lupi_features, name="w_priv_neg")
        b_priv_neg = cvx.Variable(name="bias_priv_neg")
        slack = cvx.Variable(shape=(n), name="slack")

        # Define functions for better readability
        priv_function_pos = X_priv * w_priv_pos + b_priv_pos
        priv_function_neg = X_priv * w_priv_neg + b_priv_neg

        # Combined loss of lupi function and normal slacks, scaled by two constants
        priv_loss_pos = cvx.sum(priv_function_pos)
        priv_loss_neg = cvx.sum(priv_function_neg)
        priv_loss = priv_loss_pos + priv_loss_neg
        slack_loss = cvx.sum(slack)
        loss = scaling_lupi_loss * priv_loss + slack_loss

        # L1 norm regularization of both functions with 1 scaling constant
        weight_regularization = 0.5 * (
            cvx.norm(w, 1)
            + scaling_lupi_w
            * (0.5 * cvx.norm(w_priv_pos, 1) + 0.5 * cvx.norm(w_priv_neg, 1))
        )

        constraints = [
            y - X * w - b <= epsilon + priv_function_pos + slack,
            X * w + b - y <= epsilon + priv_function_neg + slack,
            priv_function_pos >= 0,
            priv_function_neg >= 0,
            # priv_loss_pos >= 0,
            # priv_loss_neg >= 0,
            # slack_loss >= 0,
            slack >= 0,
            # loss >= 0,
        ]
        objective = cvx.Minimize(C * loss + weight_regularization)

        # Solve problem.
        solver_params = self.solver_params
        problem = cvx.Problem(objective, constraints)
        problem.solve(**solver_params)

        self.model_state = {
            "signs_pos": priv_function_pos.value > 0,
            "signs_neg": priv_function_neg.value > 0,
            "w": w.value,
            "w_priv_pos": w_priv_pos.value,
            "w_priv_neg": w_priv_neg.value,
            "b": b.value,
            "b_priv_pos": b_priv_pos.value,
            "b_priv_neg": b_priv_neg.value,
            "lupi_features": lupi_features,  # Number of lupi features in the dataset TODO: Move this somewhere else,
        }
        w_l1 = np.linalg.norm(w.value, ord=1)
        w_priv_pos_l1 = np.linalg.norm(w_priv_pos.value, ord=1)
        w_priv_neg_l1 = np.linalg.norm(w_priv_neg.value, ord=1)
        # We take the mean to combine all submodels (for priv) into a single normalization factor
        w_priv_l1 = w_priv_pos_l1 + w_priv_neg_l1
        self.constraints = {
            "priv_loss": priv_loss.value,
            "scaling_lupi_loss": scaling_lupi_loss,
            # "loss_slack": slack_loss.value,
            "loss": loss.value,
            "w_l1": w_l1,
            "w_priv_l1": w_priv_l1,
            "w_priv_pos_l1": w_priv_pos_l1,
            "w_priv_neg_l1": w_priv_neg_l1,
        }
        return self

    @property
    def solver_params(cls):
        return {"solver": "ECOS", "verbose": False}

    def predict(self, X):
        """
        Method to predict points using svm classification rule.
        We use both normal and priv. features.
        This function is mainly used for CV purposes to find the best parameters according to score.

        Parameters
        ----------
        X : numpy.ndarray
        """
        X, X_priv = split_dataset(X, self.lupi_features)
        w = self.model_state["w"]
        b = self.model_state["b"]

        y = np.dot(X, w) + b

        return y

    def score(self, X, y, **kwargs):
        prediction = self.predict(X)
        _check_reg_targets(y, prediction, None)

        score = r2_score(y, prediction)
        return score


class LUPI_Regression_Relevance_Bound(
    LUPI_Relevance_CVXProblem, Regression_Relevance_Bound
):
    @classmethod
    def generate_upper_bound_problem(
        cls,
        best_hyperparameters,
        init_constraints,
        best_model_state,
        data,
        di,
        preset_model,
        probeID=-1,
    ):
        is_priv = is_lupi_feature(
            di, data, best_model_state
        )  # Is it a lupi feature where we need additional candidate problems?

        if not is_priv:
            yield from super().generate_upper_bound_problem(
                best_hyperparameters,
                init_constraints,
                best_model_state,
                data,
                di,
                preset_model,
                probeID=probeID,
            )
        else:
            for sign, pos in product([1, -1], [True, False]):
                problem = cls(
                    di,
                    data,
                    best_hyperparameters,
                    init_constraints,
                    preset_model=preset_model,
                    best_model_state=best_model_state,
                    probeID=probeID,
                )
                problem.init_objective_UB(sign=sign, pos=pos)
                yield problem

    def _init_objective_LB_LUPI(self, **kwargs):
        self.add_constraint(
            cvx.abs(self.w_priv_pos[self.lupi_index]) <= self.feature_relevance
        )
        self.add_constraint(
            cvx.abs(self.w_priv_neg[self.lupi_index]) <= self.feature_relevance
        )

        self._objective = cvx.Minimize(self.feature_relevance)

    def _init_objective_UB_LUPI(self, pos=None, sign=None, **kwargs):
        if pos:
            self.add_constraint(
                self.feature_relevance <= sign * self.w_priv_pos[self.lupi_index]
            )
        else:
            self.add_constraint(
                self.feature_relevance <= sign * self.w_priv_neg[self.lupi_index]
            )

        self._objective = cvx.Maximize(self.feature_relevance)

    def _init_constraints(self, parameters, init_model_constraints):
        # Upper constraints from best initial model
        l1_w = init_model_constraints["w_l1"]
        self.l1_priv_w_pos = init_model_constraints["w_priv_pos_l1"]
        self.l1_priv_w_neg = init_model_constraints["w_priv_neg_l1"]
        init_loss = init_model_constraints["loss"]
        epsilon = parameters["epsilon"]
        scaling_lupi_loss = init_model_constraints["scaling_lupi_loss"]

        # New Variables
        w = cvx.Variable(shape=(self.d), name="w")
        b = cvx.Variable(name="b")
        w_priv_pos = cvx.Variable(self.d_priv, name="w_priv_pos")
        b_priv_pos = cvx.Variable(name="bias_priv_pos")
        w_priv_neg = cvx.Variable(self.d_priv, name="w_priv_neg")
        b_priv_neg = cvx.Variable(name="bias_priv_neg")
        slack = cvx.Variable(shape=(self.n))

        priv_function_pos = self.X_priv * w_priv_pos + b_priv_pos
        priv_function_neg = self.X_priv * w_priv_neg + b_priv_neg
        priv_loss = cvx.sum(priv_function_pos + priv_function_neg)
        loss = priv_loss + cvx.sum(slack)
        weight_norm = cvx.norm(w, 1)
        self.weight_norm_priv_pos = cvx.norm(w_priv_pos, 1)
        self.weight_norm_priv_neg = cvx.norm(w_priv_neg, 1)

        self.add_constraint(
            self.y - self.X * w - b <= epsilon + priv_function_pos + slack
        )
        self.add_constraint(
            self.X * w + b - self.y <= epsilon + priv_function_neg + slack
        )
        self.add_constraint(priv_function_pos >= 0)
        self.add_constraint(priv_function_neg >= 0)
        self.add_constraint(loss <= init_loss)
        self.add_constraint(slack >= 0)
        sum_norms = weight_norm + self.weight_norm_priv_pos + self.weight_norm_priv_neg
        self.add_constraint(sum_norms <= l1_w)
        # self.add_constraint(self.weight_norm_priv_pos <= self.l1_priv_w_pos)
        # self.add_constraint(self.weight_norm_priv_neg <= self.l1_priv_w_neg)

        # Save values for object use later
        self.w = w
        self.w_priv_pos = w_priv_pos
        self.w_priv_neg = w_priv_neg
        self.feature_relevance = cvx.Variable(nonneg=True, name="Feature Relevance")



from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.feature_selection.base import SelectorMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

from fri.compute import RelevanceBoundsIntervals
from fri.model.base_type import ProblemType
from fri.parameter_searcher import find_best_model

RELEVANCE_MAPPING = {0: "Irrelevant", 1: "Weak relevant", 2: "Strong relevant"}


class NotFeasibleForParameters(Exception):
    """ Problem was infeasible with the current parameter set.
    """


class FRIBase(BaseEstimator, SelectorMixin):
    def __init__(
        self,
        problem_type: ProblemType,
        random_state=None,
        n_jobs=1,
        verbose=0,
        n_param_search=30,
        n_probe_features=40,
        normalize=True,
        **kwargs,
    ):
        """

        Parameters
        ----------
        problem_type : abc.ABCMeta
        random_state : Union[mtrand.RandomState, int, None, None, None, None, None, None, None]
        n_jobs : int
        verbose : int
        n_param_search : int
        n_probe_features : int
        normalize : bool
        kwargs :

        Attributes
        ----------
        interval_ : array-like
         Feature relevance Intervals

        optim_model_ : `InitModel`
        Baseline model fitted on data

        relevance_classes_ : list(int)
        Classes of relevance encoded as int: 0 irrelevant, 1 weakly relevant, 2 strongly relevant

        relevance_classes_string_ : list(str)
        Classes of relevance encoded as string

        allrel_prediction_ : list(int)
        Relevance prediction encoded as boolean: 0 irrelevant, 1 relevant
        """

        self.n_probe_features = n_probe_features
        self.n_param_search = n_param_search

        # assert issubclass(problem_type, ProblemType)
        self.problem_type_ = problem_type(**kwargs)

        self.random_state = check_random_state(random_state)
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.normalize = normalize

        self.interval_ = None
        self.optim_model_ = None
        self.relevance_classes_ = None
        self.relevance_classes_string_ = None
        self.allrel_prediction_ = None

    def fit(self, X, y, lupi_features=0, **kwargs):
        """
        Method to fit model on data.

        Parameters
        ----------
        X : numpy.ndarray
        y : numpy.ndarray
        lupi_features : int
            Amount of features which are considered privileged information in `X`.
            The data is expected to be structured in a way that all lupi features are at the end of the set.
            For example `lupi_features=1` would denote the last column of `X` to be privileged.
        kwargs : dict
            Dictionary of additional keyword arguments depending on the `model`.

        Returns
        -------
        `FRIBase`
        """
        self.lupi_features_ = lupi_features
        self.n_samples_ = X.shape[0]
        self.n_features_ = X.shape[1] - lupi_features

        self.optim_model_, best_score = self._fit_baseline(
            X, y, lupi_features, **kwargs
        )

        data = self.problem_type_.preprocessing((X, y), lupi_features=lupi_features)
        self._relevance_bounds_computer = RelevanceBoundsIntervals(
            data,
            self.problem_type_,
            self.optim_model_,
            self.random_state,
            self.n_probe_features,
            self.n_jobs,
            self.verbose,
            normalize=self.normalize,
        )
        if lupi_features == 0:
            self.interval_, feature_classes = (
                self._relevance_bounds_computer.get_normalized_intervals()
            )
        else:
            self.interval_, feature_classes = self._relevance_bounds_computer.get_normalized_lupi_intervals(
                lupi_features=lupi_features
            )
        self._get_relevance_mask(feature_classes)

        # Return the classifier
        return self

    def _fit_baseline(self, X, y, lupi_features=0, **kwargs):

        # Preprocessing
        data = self.problem_type_.preprocessing((X, y), lupi_features=lupi_features)
        # Get predefined template for our init. model
        init_model_template = self.problem_type_.get_initmodel_template
        # Get hyperparameters which are predefined to our model template and can be seleted by user choice
        hyperparameters = self.problem_type_.get_all_parameters()
        # search_samples = len(hyperparameters) * self.n_param_search # TODO: remove this
        search_samples = self.n_param_search
        # Find an optimal, fitted model using hyperparemeter search
        optimal_model, best_score = find_best_model(
            init_model_template,
            hyperparameters,
            data,
            self.random_state,
            search_samples,
            self.n_jobs,
            self.verbose,
            lupi_features=lupi_features,
            **kwargs,
        )
        return optimal_model, best_score

    def _get_relevance_mask(self, prediction):
        """Determines relevancy using feature relevance interval values
        Parameters
        ----------
        fpr : float, optional
            false positive rate allowed under H_0
        Returns
        -------
        boolean array
            Relevancy prediction for each feature
        """

        self.relevance_classes_ = prediction
        self.relevance_classes_string_ = [RELEVANCE_MAPPING[p] for p in prediction]
        self.allrel_prediction_ = prediction > 0

        self.allrel_prediction_nonpriv_ = self.allrel_prediction_[: self.n_features_]
        self.allrel_prediction_priv_ = self.allrel_prediction_[self.n_features_ :]
        self.relevance_classes_nonpriv_ = self.relevance_classes_[: self.n_features_]
        self.relevance_classes_priv_ = self.relevance_classes_[self.n_features_ :]

        return self.allrel_prediction_

    def _n_selected_features(self):
        """

        Returns the number of selected features.
        -------

        """
        check_is_fitted(self, "allrel_prediction_")
        return sum(self.allrel_prediction_)

    def _get_support_mask(self):
        """Method for SelectorMixin
        Returns
        -------
        boolean array
        """
        return self.allrel_prediction_

    def score(self, X, y):
        """
        Using fitted model predict points for `X` and compare to truth `y`.

        Parameters
        ----------
        X : numpy.ndarray
        y : numpy.ndarray

        Returns
        -------
        Model specific score (0 is worst, 1 is best)
        """
        if self.optim_model_:
            return self.optim_model_.score(X, y)
        else:
            raise NotFittedError()

    def constrained_intervals(self, preset: dict):
        """
        Method to return relevance intervals which are constrained using preset ranges or values.

        Parameters
        ----------
        preset : dict like, {i:float} or {i:[float,float]}
            Keys denote feature index, values represent a fixed single value (float) or a range of allowed values (lower and upper bound).

            Example: To set  feature 0 to a fixed value use

            >>> preset = {0: 0.1}

            or to use the minimum relevance bound

            >>> preset[1] = self.interval_[1, 0]

        Returns
        -------
        array like
            Relevance bounds with user constraints
        """

        # Do we have intervals?
        check_is_fitted(self, "interval_")

        return self._relevance_bounds_computer.compute_multi_preset_relevance_bounds(
            preset=preset, lupi_features=self.lupi_features_
        )

    def print_interval_with_class(self):
        """

        Pretty print the relevance intervals and determined feature relevance class

        """
        output = ""
        if self.interval_ is None:
            output += "Model is not fitted."

        output += "############## Relevance bounds ##############\n"
        output += "feature: [LB -- UB], relevance class\n"
        for i in range(self.n_features_ + self.lupi_features_):
            if i == self.n_features_:
                output += "########## LUPI Relevance bounds\n"
            output += (
                f"{i:7}: [{self.interval_[i, 0]:1.1f} -- {self.interval_[i, 1]:1.1f}],"
            )
            output += f" {self.relevance_classes_string_[i]}\n"
        return output



from abc import ABC, abstractmethod

from sklearn.base import BaseEstimator


class InitModel(ABC, BaseEstimator):
    def __init__(self, **parameters):
        if parameters is None:
            parameters = {}
        self.hyperparam = parameters
        self._model_state = {}

    def _get_param_names(cls):
        return sorted(cls.hyperparameter())

    @classmethod
    @abstractmethod
    def hyperparameter(cls):
        raise NotImplementedError

    def get_params(self, deep=True):
        return self.hyperparam

    def set_params(self, **params):
        for p, value in params.items():
            self.hyperparam[p] = value
        return self

    @abstractmethod
    def fit(self, X, y, **kwargs):
        pass

    @abstractmethod
    def predict(self, X):
        pass

    @abstractmethod
    def score(self, X, y, **kwargs):
        pass

    @classmethod
    def make_scorer(self):
        return None, None

    @property
    def constraints(self):
        return self._constraints

    @constraints.setter
    def constraints(self, constraints):
        self._constraints = constraints

    @property
    def model_state(self):
        return self._model_state

    @model_state.setter
    def model_state(self, params):
        self._model_state = params

    @property
    def L1_factor(self):
        try:
            return self.constraints["w_l1"]
        except:
            raise NotImplementedError(
                "Baseline model does not provide (L1) normalization constant. Expected l1 norm of model weights (e.g. w)."
            )

    @property
    def solver_params(cls):
        return {"solver": "ECOS"}


class LUPI_InitModel(InitModel):
    @property
    def L1_factor_priv(self):
        try:
            return self.constraints["w_priv_l1"]
        except:
            raise NotImplementedError(
                "Baseline model does not provide LUPI (L1) normalization constant. Expected l1 norm of LUPI model weights (e.g. w_priv)."
            )



from itertools import product

import cvxpy as cvx
import numpy as np
from sklearn.metrics import make_scorer
from sklearn.utils import check_X_y

from fri.model.base_lupi import (
    LUPI_Relevance_CVXProblem,
    split_dataset,
    is_lupi_feature,
)
from fri.model.ordinal_regression import (
    OrdinalRegression_Relevance_Bound,
    ordinal_scores,
)
from .base_initmodel import LUPI_InitModel
from .base_type import ProblemType


class LUPI_OrdinalRegression(ProblemType):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._lupi_features = None

    @property
    def lupi_features(self):
        return self._lupi_features

    @classmethod
    def parameters(cls):
        return ["C", "scaling_lupi_w", "scaling_lupi_loss"]

    @property
    def get_initmodel_template(cls):
        return LUPI_OrdinalRegression_SVM

    @property
    def get_cvxproblem_template(cls):
        return LUPI_OrdinalRegression_Relevance_Bound

    def relax_factors(cls):
        return ["loss_slack", "w_l1_slack"]

    def preprocessing(self, data, lupi_features=None):
        X, y = data
        d = X.shape[1]

        if lupi_features is None:
            raise ValueError("Argument 'lupi_features' missing in fit() call.")
        if not isinstance(lupi_features, int):
            raise ValueError("Argument 'lupi_features' is not type int.")
        if not 0 < lupi_features < d:
            raise ValueError(
                "Argument 'lupi_features' looks wrong. We need at least 1 priviliged feature (>0) or at least one normal feature."
            )

        self._lupi_features = lupi_features

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        if np.min(y) > 0:
            print("First ordinal class has index > 0. Shifting index...")
            y = y - np.min(y)

        return X, y


class LUPI_OrdinalRegression_SVM(LUPI_InitModel):
    @classmethod
    def hyperparameter(cls):
        return ["C", "scaling_lupi_w", "scaling_lupi_loss"]

    def fit(self, X_combined, y, lupi_features=None):
        """

        Parameters
        ----------
        lupi_features : int
            Number of features in dataset which are considered privileged information (PI).
            PI features are expected to be the last features in the dataset.

        """
        if lupi_features is None:
            raise ValueError("No lupi_features argument given.")
        self.lupi_features = lupi_features
        X, X_priv = split_dataset(X_combined, lupi_features)
        (n, d) = X.shape
        self.classes_ = np.unique(y)

        # Get parameters from CV model without any feature contstraints
        C = self.hyperparam["C"]
        scaling_lupi_w = self.hyperparam["scaling_lupi_w"]
        scaling_lupi_loss = self.hyperparam["scaling_lupi_loss"]

        get_original_bin_name, n_bins = get_bin_mapping(y)
        n_boundaries = n_bins - 1

        # Initalize Variables in cvxpy
        w = cvx.Variable(shape=(d), name="w")
        b_s = cvx.Variable(shape=(n_boundaries), name="bias")

        w_priv = cvx.Variable(shape=(self.lupi_features, 2), name="w_priv")
        d_priv = cvx.Variable(shape=(2), name="bias_priv")

        slack_left = cvx.Variable(shape=(n), name="slack_left")
        slack_right = cvx.Variable(shape=(n), name="slack_right")

        def priv_function(bin, sign):
            indices = np.where(y == get_original_bin_name[bin])
            return X_priv[indices] * w_priv[:, sign] + d_priv[sign]

        # L1 norm regularization of both functions with 1 scaling constant
        priv_l1_1 = cvx.norm(w_priv[:, 0], 1)
        priv_l1_2 = cvx.norm(w_priv[:, 1], 1)
        w_priv_l1 = priv_l1_1 + priv_l1_2
        w_l1 = cvx.norm(w, 1)
        weight_regularization = 0.5 * (w_l1 + scaling_lupi_w * w_priv_l1)

        constraints = []
        loss = 0

        for left_bin in range(0, n_bins - 1):
            indices = np.where(y == get_original_bin_name[left_bin])
            constraints.append(
                X[indices] * w - b_s[left_bin] - slack_left[indices]
                <= -1 + priv_function(left_bin, 0)
            )
            constraints.append(priv_function(left_bin, 0) >= 0)
            loss += cvx.sum(priv_function(left_bin, 0))

        # Add constraints for slack into right neighboring bins
        for right_bin in range(1, n_bins):
            indices = np.where(y == get_original_bin_name[right_bin])
            constraints.append(
                X[indices] * w - b_s[right_bin - 1] - slack_right[indices]
                >= +1 - priv_function(right_bin, 1)
            )
            constraints.append(priv_function(right_bin, 1) >= 0)
            loss += cvx.sum(priv_function(right_bin, 1))

        for i_boundary in range(0, n_boundaries - 1):
            constraints.append(b_s[i_boundary] <= b_s[i_boundary + 1])

        constraints.append(slack_left >= 0)
        constraints.append(slack_right >= 0)
        loss = scaling_lupi_loss * loss + cvx.sum(slack_left + slack_right)

        objective = cvx.Minimize(C * loss + weight_regularization)

        # Solve problem.
        solver_params = self.solver_params
        problem = cvx.Problem(objective, constraints)
        problem.solve(**solver_params)

        w = w.value
        b_s = b_s.value
        self.model_state = {
            "w": w,
            "b_s": b_s,
            "w_priv": w_priv.value,
            "d_priv": d_priv.value,
            "lupi_features": lupi_features,  # Number of lupi features in the dataset TODO: Move this somewhere else
            "bin_boundaries": n_boundaries,
        }

        self.constraints = {
            "loss": loss.value,
            "w_l1": w_l1.value,
            "w_priv_l1": w_priv_l1.value,
            "priv_l1_1": priv_l1_1.value,
            "priv_l1_2": priv_l1_2.value,
        }
        return self

    def predict(self, X):

        X, X_priv = split_dataset(X, self.lupi_features)
        w = self.model_state["w"]
        b_s = self.model_state["b_s"]

        scores = np.dot(X, w.T)[np.newaxis]
        bin_thresholds = np.append(b_s, np.inf)

        # If thresholds are smaller than score the value belongs to the bigger bin
        # after subtracting we check for positive elements
        indices = np.sum(scores.T - bin_thresholds >= 0, -1)
        return self.classes_[indices]

    def score(self, X, y, error_type="mmae", return_error=False, **kwargs):

        X, y = check_X_y(X, y)

        prediction = self.predict(X)
        score = ordinal_scores(y, prediction, error_type, return_error=return_error)

        return score

    def make_scorer(self):
        # Use multiple scores for ordinal regression
        mze = make_scorer(ordinal_scores, error_type="mze")
        mae = make_scorer(ordinal_scores, error_type="mae")
        mmae = make_scorer(ordinal_scores, error_type="mmae")
        scorer = {"mze": mze, "mae": mae, "mmae": mmae}
        return scorer, "mmae"


def get_bin_mapping(y):
    """
    Get ordered unique classes and corresponding mapping from old names
    Parameters
    ----------
    y: array of discrete values (int, str)

    Returns

    -------

    """
    classes_ = np.unique(y)
    original_bins = sorted(classes_)
    n_bins = len(original_bins)
    bins = np.arange(n_bins)
    get_old_bin = dict(zip(bins, original_bins))
    return get_old_bin, n_bins


class LUPI_OrdinalRegression_Relevance_Bound(
    LUPI_Relevance_CVXProblem, OrdinalRegression_Relevance_Bound
):
    @classmethod
    def generate_lower_bound_problem(
        cls,
        best_hyperparameters,
        init_constraints,
        best_model_state,
        data,
        di,
        preset_model,
        probeID=-1,
    ):
        is_priv = is_lupi_feature(
            di, data, best_model_state
        )  # Is it a lupi feature where we need additional candidate problems?

        if not is_priv:
            yield from super().generate_lower_bound_problem(
                best_hyperparameters,
                init_constraints,
                best_model_state,
                data,
                di,
                preset_model,
                probeID=probeID,
            )
        else:
            for sign in [1, -1]:
                problem = cls(
                    di,
                    data,
                    best_hyperparameters,
                    init_constraints,
                    preset_model=preset_model,
                    best_model_state=best_model_state,
                    probeID=probeID,
                )
                problem.init_objective_LB(sign=sign)
                problem.isLowerBound = True
                yield problem

    @classmethod
    def generate_upper_bound_problem(
        cls,
        best_hyperparameters,
        init_constraints,
        best_model_state,
        data,
        di,
        preset_model,
        probeID=-1,
    ):
        is_priv = is_lupi_feature(
            di, data, best_model_state
        )  # Is it a lupi feature where we need additional candidate problems?

        if not is_priv:
            yield from super().generate_upper_bound_problem(
                best_hyperparameters,
                init_constraints,
                best_model_state,
                data,
                di,
                preset_model,
                probeID=probeID,
            )
        else:
            for sign, pos in product([1, -1], [0, 1]):
                problem = cls(
                    di,
                    data,
                    best_hyperparameters,
                    init_constraints,
                    preset_model=preset_model,
                    best_model_state=best_model_state,
                    probeID=probeID,
                )
                problem.init_objective_UB(sign=sign, pos=pos)
                yield problem

    @classmethod
    def aggregate_min_candidates(cls, min_problems_candidates):
        vals = [candidate.solved_relevance for candidate in min_problems_candidates]
        # We take the max of mins because we need the necessary contribution over all functions
        min_value = max(vals)
        return min_value

    def _init_objective_LB_LUPI(self, sign=None, bin_index=None, **kwargs):

        self.add_constraint(
            sign * self.w_priv[self.lupi_index, :] <= self.feature_relevance
        )

        self._objective = cvx.Minimize(self.feature_relevance)

    def _init_objective_UB_LUPI(self, sign=None, pos=None, **kwargs):

        self.add_constraint(
            self.feature_relevance <= sign * self.w_priv[self.lupi_index, pos]
        )

        self._objective = cvx.Maximize(self.feature_relevance)

    def _init_constraints(self, parameters, init_model_constraints):

        # Upper constraints from initial model
        init_w_l1 = init_model_constraints["w_l1"]
        init_w_priv_l1 = init_model_constraints["w_priv_l1"]
        init_priv_l1_1 = init_model_constraints["priv_l1_1"]
        init_priv_l1_2 = init_model_constraints["priv_l1_2"]
        init_loss = init_model_constraints["loss"]

        get_original_bin_name, n_bins = get_bin_mapping(self.y)
        n_boundaries = n_bins - 1

        # Initalize Variables in cvxpy
        w = cvx.Variable(shape=(self.d), name="w")
        b_s = cvx.Variable(shape=(n_boundaries), name="bias")

        w_priv = cvx.Variable(shape=(self.d_priv, 2), name="w_priv")
        d_priv = cvx.Variable(shape=(2), name="bias_priv")

        def priv_function(bin, sign):
            indices = np.where(self.y == get_original_bin_name[bin])
            return self.X_priv[indices] * w_priv[:, sign] + d_priv[sign]

        # L1 norm regularization of both functions with 1 scaling constant
        priv_l1_1 = cvx.norm(w_priv[:, 0], 1)
        priv_l1_2 = cvx.norm(w_priv[:, 1], 1)
        w_priv_l1 = priv_l1_1 + priv_l1_2
        w_l1 = cvx.norm(w, 1)

        slack_left = cvx.Variable(shape=(self.n), name="slack_left")
        slack_right = cvx.Variable(shape=(self.n), name="slack_right")

        loss = 0
        for left_bin in range(0, n_bins - 1):
            indices = np.where(self.y == get_original_bin_name[left_bin])
            self.add_constraint(
                self.X[indices] * w - b_s[left_bin]
                <= -1 + priv_function(left_bin, 0) + slack_left[indices]
            )
            self.add_constraint(priv_function(left_bin, 0) >= 0)
            loss += cvx.sum(priv_function(left_bin, 0))

        # Add constraints for slack into right neighboring bins
        for right_bin in range(1, n_bins):
            indices = np.where(self.y == get_original_bin_name[right_bin])
            self.add_constraint(
                self.X[indices] * w - b_s[right_bin - 1]
                >= +1 - priv_function(right_bin, 1) - slack_right[indices]
            )
            self.add_constraint(priv_function(right_bin, 1) >= 0)
            loss += cvx.sum(priv_function(right_bin, 1))

        loss = loss + cvx.sum(slack_left + slack_right)

        for i_boundary in range(0, n_boundaries - 1):
            self.add_constraint(b_s[i_boundary] <= b_s[i_boundary + 1])

        self.add_constraint(
            w_l1 + w_priv_l1 <= init_w_l1 + 0.5 * (init_w_priv_l1 + init_priv_l1_2)
        )
        # self.add_constraint(priv_l1_1 <= init_priv_l1_1)
        # self.add_constraint(priv_l1_2 <= init_priv_l1_2)

        self.add_constraint(slack_left >= 0)
        self.add_constraint(slack_right >= 0)
        self.add_constraint(loss <= init_loss)

        self.w = w
        self.w_priv = w_priv
        self.feature_relevance = cvx.Variable(nonneg=True, name="Feature Relevance")



import cvxpy as cvx
import numpy as np
from sklearn.metrics import make_scorer
from sklearn.utils import check_X_y

from .base_cvxproblem import Relevance_CVXProblem
from .base_initmodel import InitModel
from .base_type import ProblemType


class OrdinalRegression(ProblemType):
    @classmethod
    def parameters(cls):
        return ["C"]

    @property
    def get_initmodel_template(cls):
        return OrdinalRegression_SVM

    @property
    def get_cvxproblem_template(cls):
        return OrdinalRegression_Relevance_Bound

    def relax_factors(cls):
        return ["loss_slack", "w_l1_slack"]

    def preprocessing(self, data, **kwargs):
        X, y = data

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        if np.min(y) > 0:
            print("First ordinal class has index > 0. Shifting index...")
            y = y - np.min(y)

        return X, y


class OrdinalRegression_SVM(InitModel):
    @classmethod
    def hyperparameter(cls):
        return ["C"]

    def fit(self, X, y, **kwargs):
        (n, d) = X.shape

        C = self.hyperparam["C"]

        self.classes_ = np.unique(y)
        original_bins = sorted(self.classes_)
        n_bins = len(original_bins)
        bins = np.arange(n_bins)
        get_old_bin = dict(zip(bins, original_bins))

        w = cvx.Variable(shape=(d), name="w")
        # For ordinal regression we use two slack variables, we observe the slack in both directions
        slack_left = cvx.Variable(shape=(n), name="slack_left")
        slack_right = cvx.Variable(shape=(n), name="slack_right")
        # We have an offset for every bin boundary
        b_s = cvx.Variable(shape=(n_bins - 1), name="bias")

        objective = cvx.Minimize(cvx.norm(w, 1) + C * cvx.sum(slack_left + slack_right))
        constraints = [slack_left >= 0, slack_right >= 0]

        # Add constraints for slack into left neighboring bins
        for i in range(n_bins - 1):
            indices = np.where(y == get_old_bin[i])
            constraints.append(X[indices] * w - slack_left[indices] <= b_s[i] - 1)

        # Add constraints for slack into right neighboring bins
        for i in range(1, n_bins):
            indices = np.where(y == get_old_bin[i])
            constraints.append(X[indices] * w + slack_right[indices] >= b_s[i - 1] + 1)

        # Add explicit constraint, that all bins are ascending
        for i in range(n_bins - 2):
            constraints.append(b_s[i] <= b_s[i + 1])

        # Solve problem.
        solver_params = self.solver_params
        problem = cvx.Problem(objective, constraints)
        problem.solve(**solver_params)

        w = w.value
        b_s = b_s.value
        slack_left = np.asarray(slack_left.value).flatten()
        slack_right = np.asarray(slack_right.value).flatten()
        self.model_state = {"w": w, "b_s": b_s, "slack": (slack_left, slack_right)}

        loss = np.sum(slack_left + slack_right)
        w_l1 = np.linalg.norm(w, ord=1)
        self.constraints = {"loss": loss, "w_l1": w_l1}
        return self

    def predict(self, X):
        w = self.model_state["w"]
        b_s = self.model_state["b_s"]

        scores = np.dot(X, w.T)[np.newaxis]
        bin_thresholds = np.append(b_s, np.inf)

        # If thresholds are smaller than score the value belongs to the bigger bin
        # after subtracting we check for positive elements
        indices = np.sum(scores.T - bin_thresholds >= 0, -1)
        return self.classes_[indices]

    def score(self, X, y, error_type="mmae", return_error=False, **kwargs):

        X, y = check_X_y(X, y)

        prediction = self.predict(X)
        score = ordinal_scores(y, prediction, error_type, return_error=return_error)

        return score

    def make_scorer(self):
        # Use multiple scores for ordinal regression
        mze = make_scorer(ordinal_scores, error_type="mze")
        mae = make_scorer(ordinal_scores, error_type="mae")
        mmae = make_scorer(ordinal_scores, error_type="mmae")
        scorer = {"mze": mze, "mae": mae, "mmae": mmae}
        return scorer, "mmae"


def ordinal_scores(y, prediction, error_type, return_error=False):
    """Score function for ordinal problems.

    Parameters
    ----------
    y : target class vector
        Truth vector
    prediction : prediction class vector
        Predicted classes
    error_type : str
        Error type "mze","mae","mmae"
    return_error : bool, optional
        Return error (lower is better) or score (inverted, higher is better)

    Returns
    -------
    float
        Error or score depending on 'return_error'

    Raises
    ------
    ValueError
        When using wrong error_type
    """
    n = len(y)
    classes = np.unique(y)
    n_bins = len(classes)
    max_dist = n_bins - 1
    # If only one class available, we dont need to average
    if max_dist == 0:
        error_type = "mze"

    def mze(prediction, y):
        return np.sum(prediction != y)

    def mae(prediction, y):
        return np.sum(np.abs(prediction - y))

    # Score based on mean zero-one error
    if error_type == "mze":
        error = mze(prediction, y) / n
        score = 1 - error

    # Score based on mean absolute error
    elif error_type == "mae":
        error = mae(prediction, y) / n
        score = (max_dist - error) / max_dist

    # Score based on macro-averaged mean absolute error
    elif error_type == "mmae":
        sum = 0
        for i in range(n_bins):
            samples = y == i
            n_samples = np.sum(samples)
            if n_samples > 0:
                bin_error = mae(prediction[samples], y[samples]) / n_samples
                sum += bin_error

        error = sum / n_bins
        score = (max_dist - error) / max_dist
    else:
        raise ValueError("error_type {} not available!'".format(error_type))

    if return_error:
        return error
    else:
        return score


class OrdinalRegression_Relevance_Bound(Relevance_CVXProblem):
    def init_objective_UB(self, sign=None, **kwargs):

        self.add_constraint(
            self.feature_relevance <= sign * self.w[self.current_feature]
        )
        self._objective = cvx.Maximize(self.feature_relevance)

    def init_objective_LB(self, **kwargs):
        self.add_constraint(
            cvx.abs(self.w[self.current_feature]) <= self.feature_relevance
        )
        self._objective = cvx.Minimize(self.feature_relevance)

    def _init_constraints(self, parameters, init_model_constraints):

        n_bins = len(np.unique(self.y))

        # Upper constraints from initial model
        l1_w = init_model_constraints["w_l1"]
        init_loss = init_model_constraints["loss"]

        C = parameters["C"]

        # New Variables
        self.w = cvx.Variable(shape=(self.d), name="w")

        # For ordinal regression we use two slack variables, we observe the slack in both directions
        self.slack_left = cvx.Variable(shape=(self.n), name="slack_left", nonneg=True)
        self.slack_right = cvx.Variable(shape=(self.n), name="slack_right", nonneg=True)

        # We have an offset for every bin boundary
        self.b_s = cvx.Variable(shape=(n_bins - 1), name="bias")

        # New Constraints
        self.loss = cvx.sum(self.slack_left + self.slack_right)
        self.weight_norm = cvx.norm(self.w, 1)

        for i in range(n_bins - 1):
            indices = np.where(self.y == i)
            self.add_constraint(
                self.X[indices] * self.w - self.slack_left[indices] <= self.b_s[i] - 1
            )

        for i in range(1, n_bins):
            indices = np.where(self.y == i)
            self.add_constraint(
                self.X[indices] * self.w + self.slack_right[indices]
                >= self.b_s[i - 1] + 1
            )

        for i in range(n_bins - 2):
            self.add_constraint(self.b_s[i] <= self.b_s[i + 1])

        self.add_constraint(self.weight_norm <= l1_w)
        self.add_constraint(C * self.loss <= C * init_loss)

        self.feature_relevance = cvx.Variable(nonneg=True, name="Feature Relevance")



import numpy as np
import pytest
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_random_state

from fri import FRI, ProblemName
from fri import genRegressionData, genClassificationData, genOrdinalRegressionData


@pytest.fixture(scope="function")
def randomstate():
    return check_random_state(1337)


@pytest.mark.parametrize("n_weak", [0, 2, 3])
@pytest.mark.parametrize("n_strong", [0, 1, 2])
@pytest.mark.parametrize("problem", ["regression", "classification", "ordreg"])
def test_model(problem, n_strong, n_weak, randomstate):
    n_samples = 300
    n_features = 8

    if problem is "regression":
        gen = genRegressionData
        model = FRI(ProblemName.REGRESSION, random_state=randomstate, verbose=1)

    elif problem is "classification":
        gen = genClassificationData
        model = FRI(ProblemName.CLASSIFICATION, random_state=randomstate, verbose=1)
    elif problem is "ordreg":
        gen = genOrdinalRegressionData
        model = FRI(ProblemName.ORDINALREGRESSION, random_state=randomstate, verbose=1)

    if n_strong + n_weak == 0:
        with pytest.raises(ValueError):
            gen(
                n_samples=n_samples,
                n_features=n_features,
                n_redundant=n_weak,
                n_strel=n_strong,
                n_repeated=0,
                random_state=randomstate,
            )

    else:
        data = gen(
            n_samples=n_samples,
            n_features=n_features,
            n_redundant=n_weak,
            n_strel=n_strong,
            n_repeated=0,
            noise=0,
            random_state=randomstate,
        )

        X_orig, y = data
        X_orig = StandardScaler().fit(X_orig).transform(X_orig)
        X = X_orig

        model.fit(X, y)

        # Check the interval output
        interval = model.interval_
        assert len(model.allrel_prediction_) == X.shape[1]
        assert len(interval) == X.shape[1]

        # Check the score which should be good
        if problem is not "ordreg":
            assert model.score(X[:30], y[:30]) >= 0.8

        n_f = n_strong + n_weak  # Number of relevant features

        # Check how many are selected
        selected = model._n_selected_features()
        # we allow one more false positive
        print(model._get_support_mask())
        print(model.interval_)
        print(model.print_interval_with_class())
        assert n_f == selected

        # Check if all relevant features are selected
        truth = np.ones(n_f)
        assert all(model._get_support_mask()[:n_f] == truth)


def test_multiprocessing(randomstate):
    data = genClassificationData(
        n_samples=100, n_features=3, n_redundant=2, n_strel=1, random_state=randomstate
    )

    X_orig, y = data
    X = StandardScaler().fit(X_orig).transform(X_orig)

    model = FRI(ProblemName.CLASSIFICATION, random_state=randomstate, n_jobs=1)
    model.fit(X, y)

    model = FRI(ProblemName.CLASSIFICATION, random_state=randomstate, n_jobs=2)
    model.fit(X, y)

    model = FRI(ProblemName.CLASSIFICATION, random_state=randomstate, n_jobs=-1)
    model.fit(X, y)


def test_nonbinaryclasses(randomstate):
    n = 90
    d = 2
    X = randomstate.rand(n, d)
    firstclass = [1] * 30
    secondclass = [2] * 30
    thirdclass = [3] * 30
    y = np.array([firstclass, secondclass, thirdclass]).ravel()

    fri = FRI(ProblemName.CLASSIFICATION)
    with pytest.raises(ValueError):
        fri.fit(X, y)



from abc import abstractmethod

from .base_cvxproblem import Relevance_CVXProblem


class LUPI_Relevance_CVXProblem(Relevance_CVXProblem):
    def __init__(
        self,
        current_feature: int,
        data: tuple,
        hyperparameters,
        best_model_constraints,
        preset_model=None,
        best_model_state=None,
        probeID=-1,
    ) -> None:
        super().__init__(
            current_feature,
            data,
            hyperparameters,
            best_model_constraints,
            preset_model,
            best_model_state,
            probeID,
        )

    def preprocessing_data(self, data, best_model_state):
        lupi_features = best_model_state["lupi_features"]
        X_combined, y = data
        X, X_priv = split_dataset(X_combined, lupi_features)
        self.X_priv = X_priv
        super().preprocessing_data((X, y), best_model_state)

        assert lupi_features == X_priv.shape[1]
        self.d_priv = lupi_features
        # LUPI model, we need to offset the index
        self.lupi_index = self.current_feature - self.d
        if self.lupi_index >= 0:
            self.isPriv = True
        else:
            self.isPriv = False

    def init_objective_UB(self, **kwargs):
        # We have two models basically with different indexes
        if self.isPriv:
            self._init_objective_UB_LUPI(**kwargs)
        else:
            # We call sibling class of our lupi class, which is the normal problem
            super().init_objective_UB(**kwargs)

    def init_objective_LB(self, **kwargs):
        # We have two models basically with different indexes
        if self.isPriv:
            self._init_objective_LB_LUPI(**kwargs)
        else:
            # We call sibling class of our lupi class, which is the normal problem
            super().init_objective_LB(**kwargs)

    @abstractmethod
    def _init_objective_LB_LUPI(self, **kwargs):
        pass

    @abstractmethod
    def _init_objective_UB_LUPI(self, **kwargs):
        pass


def split_dataset(X_combined, lupi_features):
    assert X_combined.shape[1] > lupi_features
    X = X_combined[:, :-lupi_features]
    X_priv = X_combined[:, -lupi_features:]
    return X, X_priv


def is_lupi_feature(di, data, best_model_state):
    lupi_features = best_model_state["lupi_features"]
    X_combined, _ = data
    d = X_combined.shape[1] - lupi_features
    lupi_index = di - d
    return lupi_index >= 0



import pytest
from sklearn import linear_model
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.svm import LinearSVR
from sklearn.utils import check_random_state
from sklearn.utils.testing import assert_equal

from fri import ProblemName
from fri import quick_generate
from fri.toydata import (
    genClassificationData,
    genRegressionData,
    genOrdinalRegressionData,
)


@pytest.fixture(scope="function")
def random_state():
    return check_random_state(1337)


@pytest.mark.parametrize("problem", list(ProblemName))
def test_quick_generate(problem):
    data = quick_generate(problem)
    assert (
        len(data) == 2 or len(data) == 3
    )  # Check if two-tuple or three-tuple depending on if we have lupi data or not


@pytest.mark.parametrize("n_samples", [2, 100, 10000])
@pytest.mark.parametrize("n_dim", [1, 5, 30, 100, 1000])
def test_shape(n_samples, n_dim):
    X, y = genClassificationData(n_samples=n_samples, n_features=n_dim)

    # Equal length
    assert_equal(len(X), len(y))
    # Correct parameters
    assert_equal(len(X), n_samples)
    assert_equal(X.shape[1], n_dim)


@pytest.mark.parametrize(
    "wrong_param",
    [
        {"n_samples": 0},
        {"n_features": 0},
        {"n_samples": -1},
        {"n_features": -1},
        {"n_features": 2, "n_strel": 3},
        {"n_features": 2, "n_redundant": 3},
        {"n_features": 2, "n_repeated": 3},
    ],
)
def test_wrong_values(wrong_param):
    with pytest.raises(ValueError) as exc:
        genClassificationData(**wrong_param)


@pytest.mark.parametrize("strong", [0, 1, 2, 20, 50])
@pytest.mark.parametrize("weak", [0, 1, 2, 20])
@pytest.mark.parametrize("repeated", [0, 1, 2, 5])
@pytest.mark.parametrize("flip_y", [0, 0.1, 1])
def test_genClassification(strong, weak, repeated, flip_y):
    n_samples = 10
    n_features = 100
    args = {
        "n_samples": n_samples,
        "n_features": n_features,
        "n_strel": strong,
        "n_redundant": weak,
        "n_repeated": repeated,
        "flip_y": flip_y,
    }

    gen = genClassificationData
    if flip_y == 1:
        with pytest.raises(ValueError):
            X, y = gen(**args)
        return

    if strong == 0 and weak < 2:
        with pytest.raises(ValueError):
            X, y = gen(**args)
        return

    X, y = gen(**args)

    # Equal length
    assert_equal(len(X), len(y))
    # Correct parameters
    assert_equal(len(X), n_samples)
    assert_equal(X.shape[1], n_features)


@pytest.mark.parametrize("strong", [0, 1, 2, 20, 50])
@pytest.mark.parametrize("weak", [0, 1, 2, 20])
@pytest.mark.parametrize("repeated", [0, 1, 2, 5])
@pytest.mark.parametrize("noise", [0, 0.5, 1, 10])
def test_genRegression(strong, weak, repeated, noise):
    n_samples = 10
    n_features = 100
    args = {
        "n_samples": n_samples,
        "n_features": n_features,
        "n_strel": strong,
        "n_redundant": weak,
        "n_repeated": repeated,
        "noise": noise,
    }

    gen = genRegressionData

    if strong == 0 and weak < 2:
        with pytest.raises(ValueError):
            X, y = gen(**args)
        return

    X, y = gen(**args)

    # Equal length
    assert_equal(len(X), len(y))
    # Correct parameters
    assert_equal(len(X), n_samples)
    assert_equal(X.shape[1], n_features)


@pytest.mark.parametrize("strong", [0, 1, 2, 20, 50])
@pytest.mark.parametrize("n_samples", [100, 101])
@pytest.mark.parametrize("weak", [0, 1, 2, 20])
@pytest.mark.parametrize("repeated", [0, 1, 2, 5])
@pytest.mark.parametrize("noise", [0, 0.5, 1, 10])
@pytest.mark.parametrize("bins", [3, 4, 10, 20])
def test_genOrdinalRegression(strong, weak, repeated, noise, bins, n_samples):
    n_features = 100
    args = {
        "n_samples": n_samples,
        "n_features": n_features,
        "n_strel": strong,
        "n_redundant": weak,
        "n_repeated": repeated,
        "noise": noise,
        "n_target_bins": bins,
    }

    gen = genOrdinalRegressionData

    if strong == 0 and weak < 2:
        with pytest.raises(ValueError):
            X, y = gen(**args)
        return

    X, y = gen(**args)

    # Equal length
    assert_equal(len(X), len(y))
    # Correct parameters
    assert_equal(len(X), n_samples)
    assert_equal(X.shape[1], n_features)


def test_data_truth():
    n = 100
    d = 10
    strRel = 2
    generator = check_random_state(1337)
    X, Y = genRegressionData(
        n_samples=n,
        n_features=d,
        n_redundant=0,
        n_strel=strRel,
        n_repeated=0,
        random_state=generator,
        noise=0,
    )
    X = StandardScaler().fit_transform(X)
    X_train, X_test, y_train, y_test = train_test_split(X, Y, random_state=generator)

    linsvr = LinearSVR()
    linsvr.fit(X_train, y_train)
    pred = linsvr.predict(X_test)
    r2 = r2_score(y_test, pred)

    assert r2 > 0.9


def test_data_noise(random_state):
    n = 100
    d = 10
    strRel = 5

    X, Y = genRegressionData(
        n_samples=n,
        n_features=d,
        n_redundant=0,
        n_strel=strRel,
        n_repeated=0,
        random_state=random_state,
        noise=100,
    )
    X = StandardScaler().fit_transform(X)
    X_train, X_test, y_train, y_test = train_test_split(
        X, Y, test_size=0.33, random_state=random_state
    )
    reg = linear_model.LinearRegression(normalize=True)
    reg.fit(X_train, y_train)

    testscore = reg.score(X_test, y_test)
    assert testscore < 0.55


@pytest.mark.parametrize("problem", ["regression", "classification"])
@pytest.mark.parametrize(
    "partition",
    [
        [2, 2, 3],
        [5, 10, 3],
        pytest.param([2, 0, 2], marks=pytest.mark.xfail(raises=ValueError)),
        pytest.param([2, 1, 2], marks=pytest.mark.xfail(raises=ValueError)),
    ],
)
def test_partition(problem, partition):
    n_samples = 10
    n_features = 100
    strong = 5
    weak = sum(partition)
    repeated = 0

    args = {
        "n_samples": n_samples,
        "n_features": n_features,
        "n_strel": strong,
        "n_redundant": weak,
        "n_repeated": repeated,
        "partition": partition,
    }

    if problem == "regression":
        gen = genRegressionData
    else:
        gen = genClassificationData

    X, y = gen(**args)

    # Equal length
    assert_equal(len(X), len(y))
    # Correct parameters
    assert_equal(len(X), n_samples)
    assert_equal(X.shape[1], n_features)



import numpy as np
from numpy.random import RandomState
from sklearn.utils import check_random_state
from sklearn.utils import shuffle


def _combFeat(n, size, strRelFeat, randomstate):
    # Split each strongly relevant feature into linear combination of it
    weakFeats = np.tile(strRelFeat, (size, 1)).T
    weakFeats = randomstate.uniform(low=1, high=2, size=size) + weakFeats
    return weakFeats


def _dummyFeat(n, randomstate):
    return randomstate.randn(n)


def _repeatFeat(feats, i, randomstate):
    i_pick = randomstate.choice(i)
    return feats[:, i_pick]


def _checkParam(
    n_samples: int = 100,
    n_redundant: int = 0,
    n_strel: int = 1,
    n_repeated: int = 0,
    n_features=1,
    flip_y: float = 0,
    noise: float = 1,
    partition=None,
    **kwargs,
):
    if not 1 < n_samples:
        raise ValueError("We need at least 2 samples.")
    if not 0 < n_features:
        raise ValueError("We need at least one feature.")
    if not 0 <= flip_y < 1:
        raise ValueError("Flip percentage has to be between 0 and 1.")
    if not n_redundant + n_repeated + n_strel <= n_features:
        raise ValueError("Inconsistent number of features")
    if n_strel + n_redundant < 1:
        raise ValueError("No informative features.")
    if n_strel == 0 and n_redundant < 2:
        raise ValueError("We need more than 1 redundant feature.")
    if partition is not None:
        if sum(partition) != n_redundant:
            raise ValueError(
                "Sum of partition values should yield number of redundant features."
            )
        if 0 in partition or 1 in partition:
            raise ValueError(
                "Subset defined in Partition needs at least 2 features. 0 and 1 is not allowed."
            )


def _fillVariableSpace(
    X_informative,
    random_state: RandomState,
    n_features: int = 2,
    n_redundant: int = 0,
    n_strel: int = 1,
    n_repeated: int = 0,
    partition=None,
):
    n_samples = len(X_informative)

    if partition is not None:
        assert n_redundant == np.sum(partition)

    # Create dummy array
    X = np.zeros((int(n_samples), int(n_features)))

    # Add strongly relevant
    X[:, :n_strel] = X_informative[:, :n_strel]

    # Save strongly relevant used in creation of weakly ones
    holdout = X_informative[:, n_strel:]

    i = n_strel
    pi = 0
    for x in range(holdout.shape[1]):
        size = partition[pi]
        X[:, i : i + size] = _combFeat(n_samples, size, holdout[:, x], random_state)
        i += size
        pi += 1

    for x in range(n_repeated):
        X[:, i] = _repeatFeat(X[:, :i], i, random_state)
        i += 1
    for x in range(n_features - i):
        X[:, i] = _dummyFeat(n_samples, random_state)
        i += 1

    return X


def generate_binary_classification_problem(
    n_samples: int, features: int, random_state: RandomState = None, data_range=1
):
    """ Generate data uniformly distributed in a square and perfectly separated by the hyperplane given by normal_vector and b.
    Keyword arguments:
    n_samples -- number of samples required (default 100)
    n_features -- number of features required
    normal_vector -- the normal vector of the separating hyperplane
    data_range -- data is distributed between -data_range and data_range (default 10)

    """
    random_state = check_random_state(random_state)

    data = random_state.normal(size=(n_samples, features), scale=data_range)
    labels = np.sum(data, 1) > 0
    labels = labels.astype(int)
    labels[labels == 0] = -1
    return data, labels


def genClassificationData(
    n_samples: int = 100,
    n_features: int = 2,
    n_redundant: int = 0,
    n_strel: int = 1,
    n_repeated: int = 0,
    noise: float = 0.1,
    flip_y: float = 0,
    random_state: object = None,
    partition=None,
):
    """Generate synthetic classification data
    
    Parameters
    ----------
    n_samples : int, optional
        Number of samples
    n_features : int, optional
        Number of features
    n_redundant : int, optional
        Number of features which are part of redundant subsets (weakly relevant)
    n_strel : int, optional
        Number of features which are mandatory for the underlying model (strongly relevant)
    n_repeated : int, optional
        Number of features which are clones of existing ones. 
    noise : float
        Added gaussian noise to data. Parameter scales Std of normal distribution.
    flip_y : float, optional
        Ratio of samples randomly switched to wrong class.
    random_state : object, optional
        Randomstate object used for generation.
    
    Returns
    -------
    X : array of shape [n_samples, n_features]
        The generated samples.
    y : array of shape [n_samples]
        The output classes.
    
    Raises
    ------
    ValueError
        Description
    ValueError
    Wrong parameters for specified amonut of features/samples.
    

    Examples
    ---------
    >>> X,y = genClassificationData(n_samples=200)
    Generating dataset with d=2,n=200,strongly=1,weakly=0, partition of weakly=None
    >>> X.shape
    (200, 2)
    >>> y.shape
    (200,)
    """
    _checkParam(**locals())
    random_state = check_random_state(random_state)

    X = np.zeros((n_samples, n_features))

    # Find partitions which defíne the weakly relevant subsets
    if partition is None and n_redundant > 0:
        partition = [n_redundant]
        part_size = 1
    elif partition is not None:
        part_size = len(partition)
    else:
        part_size = 0

    X_informative, Y = generate_binary_classification_problem(
        n_samples, n_strel + part_size, random_state
    )

    X = _fillVariableSpace(
        X_informative,
        random_state,
        n_features=n_features,
        n_redundant=n_redundant,
        n_strel=n_strel,
        n_repeated=n_repeated,
        partition=partition,
    )
    # Add target noise
    if flip_y > 0:
        n_flip = int(flip_y * n_samples)
        Y[random_state.choice(n_samples, n_flip)] *= -1

    # Add gaussian noise to data
    X = X + random_state.normal(size=(n_samples, n_features), scale=noise / X.std())

    return X, Y


def genRegressionData(
    n_samples: int = 100,
    n_features: int = 2,
    n_redundant: int = 0,
    n_strel: int = 1,
    n_repeated: int = 0,
    noise: float = 0.0,
    random_state: object = None,
    partition=None,
) -> object:
    """Generate synthetic regression data
    
    Parameters
    ----------
    n_samples : int, optional
        Number of samples
    n_features : int, optional
        Number of features
    n_redundant : int, optional
        Number of features which are part of redundant subsets (weakly relevant)
    n_strel : int, optional
        Number of features which are mandatory for the underlying model (strongly relevant)
    n_repeated : int, optional
        Number of features which are clones of existing ones. 
    noise : float, optional
        Noise of the created samples around ground truth.
    random_state : object, optional
        Randomstate object used for generation.

    Returns
    -------
    X : array of shape [n_samples, n_features]
        The generated samples.
    y : array of shape [n_samples]
        The output values (target).
    
    Raises
    ------
    ValueError
    Wrong parameters for specified amonut of features/samples.
    """

    _checkParam(**locals())
    random_state = check_random_state(random_state)

    # Find partitions which defíne the weakly relevant subsets
    if partition is None and n_redundant > 0:
        partition = [n_redundant]
        part_size = 1
    elif partition is not None:
        part_size = len(partition)
    else:
        part_size = 0

    n_informative = n_strel + part_size

    X = random_state.randn(n_samples, n_informative)
    ground_truth = np.zeros((n_informative, 1))
    ground_truth[:n_informative, :] = 0.3
    bias = 0

    y = np.dot(X, ground_truth) + bias

    # Add noise
    if noise > 0.0:
        y += random_state.normal(scale=noise, size=y.shape)

    X = _fillVariableSpace(
        X,
        random_state,
        n_features=n_features,
        n_redundant=n_redundant,
        n_strel=n_strel,
        n_repeated=n_repeated,
        partition=partition,
    )
    y = np.squeeze(y)
    return X, y


def genOrdinalRegressionData(
    n_samples: int = 100,
    n_features: int = 2,
    n_redundant: int = 0,
    n_strel: int = 1,
    n_repeated: int = 0,
    noise: float = 0.0,
    random_state: object = None,
    partition=None,
    n_target_bins: int = 3,
):
    """
    Generate ordinal regression data

    Parameters
    ----------
    n_samples : int, optional
        Number of samples
    n_features : int, optional
        Number of features
    n_redundant : int, optional
        Number of features which are part of redundant subsets (weakly relevant)
    n_strel : int, optional
        Number of features which are mandatory for the underlying model (strongly relevant)
    n_repeated : int, optional
        Number of features which are clones of existing ones.
    noise : float, optional
        Noise of the created samples around ground truth.
    random_state : object, optional
        Randomstate object used for generation.
    n_target_bins : int, optional
        Number of bins in which the regressional target variable is split to form the ordinal classes

    Returns
    -------
    X : array of shape [n_samples, n_features]
        The generated samples.
    y : array of shape [n_samples]
        The output values (target).

    Raises
    ------
    ValueError
    Wrong parameters for specified amonut of features/samples.
    """

    _checkParam(**locals())
    random_state = check_random_state(random_state)

    if not n_target_bins > 1:
        raise ValueError("At least 2 target bins needed")

    # Use normal regression data as starting point
    X_regression, Y_regression = genRegressionData(
        n_samples=int(n_samples),
        n_features=int(n_features),
        n_redundant=int(n_redundant),
        n_strel=int(n_strel),
        n_repeated=int(n_repeated),
        noise=0,
        random_state=random_state,
        partition=partition,
    )

    bin_size = int(np.floor(n_samples / n_target_bins))
    rest = int(n_samples - (bin_size * n_target_bins))

    # Sort the target values and rearange the data accordingly
    sort_indices = np.argsort(Y_regression)
    X = X_regression[sort_indices]
    Y = Y_regression[sort_indices]

    # Assign ordinal classes as target values
    for i in range(n_target_bins):
        Y[bin_size * i : bin_size * (i + 1)] = i

    # Put non divisable rest into last bin
    if rest > 0:
        Y[-rest:] = n_target_bins - 1

    X, Y = shuffle(X, Y, random_state=random_state)

    # Add gaussian noise to data
    X = X + random_state.normal(size=(n_samples, n_features), scale=noise)

    return X, Y



import logging

# noinspection PyUnresolvedReferences
from fri._version import __version__

logging.basicConfig(level=logging.INFO)
from enum import Enum
import fri.model


class ProblemName(Enum):
    """
    Enum which contains usable models for which feature relevance intervals can be computed in :func:`~FRI`.
    """

    CLASSIFICATION = fri.model.Classification
    REGRESSION = fri.model.Regression
    ORDINALREGRESSION = fri.model.OrdinalRegression
    LUPI_CLASSIFICATION = fri.model.LUPI_Classification
    LUPI_REGRESSION = fri.model.LUPI_Regression
    LUPI_ORDREGRESSION = fri.model.LUPI_OrdinalRegression


NORMAL_MODELS = [
    ProblemName.CLASSIFICATION,
    ProblemName.REGRESSION,
    ProblemName.ORDINALREGRESSION,
]
LUPI_MODELS = [
    ProblemName.LUPI_CLASSIFICATION,
    ProblemName.LUPI_REGRESSION,
    ProblemName.LUPI_ORDREGRESSION,
]

from fri.toydata import (
    genRegressionData,
    genClassificationData,
    genOrdinalRegressionData,
    quick_generate,
    genLupiData,
)
from fri.main import FRIBase
from fri.plot import plot_intervals


class FRI(FRIBase):
    def __init__(
        self,
        problemName: object,
        random_state: object = None,
        n_jobs: object = 1,
        verbose: object = 0,
        n_param_search: object = 10,
        n_probe_features: object = 20,
        slack_regularization: object = 0.001,
        slack_loss: object = 0.001,
        normalize: object = True,
        **kwargs,
    ):
        """
        Main class to use `FRI` in programattic fashion following the scikit-learn paradigm.

        Parameters
        ----------
        problemName: `ProblemName` or str
            Type of Problem as enum value or explicit string (e.g. "classification")
        random_state: object or int
            Random state object or int
        n_jobs: int or None
            Number of threads or -1 for automatic.
        verbose: int
            Verbosity if > 0
        n_param_search: int
            Number of parameter samples in random search for hyperparameters.
        n_probe_features: int
            Number of probes to generate to improve feature selection.
        slack_regularization: float
            Allow deviation from optimal L1 norm.
        slack_loss: float
            Allow deviation of loss.
        normalize: boolean
            Normalize relevace bounds to range of [0,1] depending on L1 norm.

        """
        if isinstance(problemName, ProblemName):
            problemtype = problemName.value
        else:
            if problemName == "classification" or problemName == "class":
                problemtype = ProblemName.CLASSIFICATION
            elif problemName == "regression" or problemName == "reg":
                problemtype = ProblemName.REGRESSION
            elif problemName == "ordinalregression" or problemName == "ordreg":
                problemtype = ProblemName.ORDINALREGRESSION
            elif problemName == "lupi_classification" or problemName == "lupi_class":
                problemtype = ProblemName.LUPI_CLASSIFICATION

        if problemtype is None:
            names = [enum.name.lower() for enum in ProblemName]
            print(
                f"Parameter 'problemName' was not recognized or unset. Try one of {names}."
            )
        else:
            super().__init__(
                problemtype,
                random_state=random_state,
                n_jobs=n_jobs,
                verbose=verbose,
                n_param_search=n_param_search,
                n_probe_features=n_probe_features,
                w_l1_slack=slack_regularization,
                loss_slack=slack_loss,
                normalize=normalize,
                **kwargs,
            )


__all__ = [
    "genRegressionData",
    "genClassificationData",
    "genOrdinalRegressionData",
    "quick_generate",
    "plot_intervals",
    "ProblemName",
    "FRI",
    "LUPI_MODELS",
    "NORMAL_MODELS",
    "genLupiData",
]



import numpy as np
import pytest
from sklearn.utils import check_random_state

from fri import FRI, NORMAL_MODELS, LUPI_MODELS
from fri import quick_generate


@pytest.fixture(scope="function")
def random_state():
    return check_random_state(1337)


@pytest.fixture()
def data(prob):
    return quick_generate(prob)


@pytest.mark.parametrize("problem", NORMAL_MODELS)
def test_normal_model(problem, random_state):
    model = FRI(problem, random_state=random_state)

    X, y = quick_generate(problem, random_state=random_state)

    model.fit(X, y)

    assert len(model.allrel_prediction_) == X.shape[1]


@pytest.mark.parametrize("problem", LUPI_MODELS)
def test_lupi_model(problem, random_state):
    model = FRI(problem, random_state=random_state)

    X, X_p, y = quick_generate(problem, random_state=random_state)
    combined = np.hstack([X, X_p])
    model.fit(combined, y, lupi_features=X.shape[1])

    assert len(model.allrel_prediction_) == combined.shape[1]



import numpy as np
from sklearn.utils import check_random_state

from fri import ProblemName
from .gen_data import _fillVariableSpace


def _checkLupiParam(
    problemName, lupiType, n_strel, n_weakrel, n_priv_weakrel, partition, partition_priv
):
    """
        Checks if the parameters supplied to the genLupiData() function are okay.

        Parameters
        ----------
        problemName : Str
            Must be one of ['classification', 'regression', 'ordinalRegression']
        lupiType : Str
            Must be one of ['cleanLabels', 'cleanFeatures']
        n_strel : int
            Stands for the number of strongly relevant features to generate in genLupiData()
            Must be greater than 0
        n_weakrel : int
            Must be equal to the length of the partition list
        n_priv_weakrel : int
            Must be equal to the length of the partition_priv list
        partition : list of int
            The length of the list must be equal to n_weakrel
        partition_priv : list of int
            The length of the list must be equal to n_priv_weakrel
    """

    if type(problemName) is not ProblemName:
        raise ValueError("Not of Type ProblemName")
    if lupiType not in ["cleanLabels", "cleanFeatures"]:
        raise ValueError(
            "The lupiType parameter must be a string out of ['cleanLabels', 'cleanFeatures']."
        )
    if n_strel < 1:
        raise ValueError(
            "At least one strongly relevant feature is necessary (Parmeter 'n_strel' must be greater than 0)."
        )
    if partition is not None:
        if sum(partition) != n_weakrel:
            raise ValueError(
                "The sum over the entries in the partition list must be equal to the parameter 'n_weakrel'."
            )
        if 0 in partition or 1 in partition:
            raise ValueError(
                "The entries in the partition list must be greater or equal to 2."
            )
    if partition_priv is not None:
        if sum(partition_priv) != n_priv_weakrel:
            raise ValueError(
                "The sum over the entries in the partition_priv list must be equal to the parameter 'n_priv_weakrel'."
            )
        if 0 in partition_priv or 1 in partition_priv:
            raise ValueError(
                "The entries in the partition_priv list must be greater or equal to 2."
            )
    if lupiType == "cleanLabels" and n_priv_weakrel > 0:
        raise ValueError(
            "The 'cleanLabels' data has only one strongly relevant feature by nature, this can be repeated ('n_priv_repeated'),"
            "or useless information can be added ('n_priv_irrel') but it can not be weakend => n_priv_weakrel hast to be 0."
        )


# def _genWeakFeatures(n_weakrel, X, random_state, partition):
#     """
#         Generate n_weakrel features out of the strRelFeature
#
#         Parameters
#         ----------
#         n_weakrel : int
#             Number of weakly relevant feature to be generated
#         X : array of shape [n_samples, n_features]
#             Contains the data out of which the weakly relevant features are created
#         random_state : Random State object
#             Used to generate the samples
#         partition : list of int
#             Used to define how many weak features are calculated from the same strong feature
#             The sum of the entries in the partition list must be equal to n_weakrel
#
#
#         Returns
#         ----------
#         X_weakrel : array of shape [n_samples, n_weakrel]
#             Contains the data of the generated weak relevant features
#     """
#
#     X_weakrel = np.zeros([X.shape[0], n_weakrel])
#
#     if partition is None:
#         for i in range(n_weakrel):
#             X_weakrel[:, i] = X[
#                 :, random_state.choice(X.shape[1])
#             ] + random_state.normal(loc=0, scale=1, size=1)
#     else:
#         idx = 0
#         for j in range(len(partition)):
#             X_weakrel[:, idx : idx + partition[j]] = np.tile(
#                 X[:, random_state.choice(X.shape[1])], (partition[j], 1)
#             ).T + random_state.normal(loc=0, scale=1, size=partition[j])
#             idx += partition[j]
#
#     return X_weakrel
#
#
# def _genRepeatedFeatures(n_repeated, X, random_state):
#     """
#         Generate repeated features by picking a random existing feature out of X
#
#         Parameters
#         ----------
#         n_repeated : int
#             Number of repeated features to create
#         X : array of shape [n_samples, n_features]
#             Contains the data of which the repeated features are picked
#         random_state : Random State object
#             Used to randomly pick a feature out of X
#     """
#
#     X_repeated = np.zeros([X.shape[0], n_repeated])
#     for i in range(n_repeated):
#         X_repeated[:, i] = X[:, random_state.choice(X.shape[1])]
#
#     return X_repeated


def genLupiData(
    problemName: ProblemName,
    n_samples: int = 100,
    random_state: object = None,
    n_ordinal_bins: int = 3,
    n_strel: int = 1,
    n_weakrel: int = 0,
    n_repeated: int = 0,
    n_irrel: int = 0,
    label_noise=0.0,
):
    """
            Generate Lupi Data for Classification, Regression and Ordinal Regression Problems

            Parameters
            ----------
            problemName : ProblemName
                Defines the type of y-values of the problem. Example `ProblemName.CLASSIFICATION`.
            n_samples : int, optional
                Number of samples
            random_state : object, optional
                Randomstate object used for generation.
            n_ordinal_bins : int, optional
                Number of bins in which the regressional target variable is split to form the ordinal classes,
                Only has an effect if problemType == 'ordinalRegression'
            n_strel : int, optional
                Number of features which are mandatory for the underlying model (strongly relevant)
            n_weakrel : int, optional
                Number of weakly relevant features
            n_repeated : int, optional
                Number of features which are clones of existing ones.
            n_irrel : int, optional
                Number of features which are irrelevant to the underlying model
            label_noise: float, optional
                Percentage of labels which get permutated.


            Returns
            -------
            X : array of shape [n_samples, (n_strel + n_weakrel + n_repeated + n_irrel)]
                The generated samples
            X_priv : array with same shape as X
                The generated privileged samples
            y : array of shape [n_samples]
                The generated target values
                In case of problemType == 'classification' : values are in [0,1]
                In case of problemType == 'regression' : values are continious
                In case of problemType == 'ordinalRegression' : values are in [0, n_ordinal_bins]


        """

    random_state = check_random_state(random_state)
    n_informative = n_strel + (n_weakrel > 0)

    # Create truth (prototype) vector which contains true feature contributions
    # We enforce minimum of 0.1 to circumvent problems when testing for relevance
    w = random_state.uniform(low=0.5, high=1, size=n_informative)
    X_informative = random_state.normal(size=(n_samples, n_informative))
    scores = np.dot(X_informative, w)

    n_features = n_strel + n_weakrel + n_repeated + n_irrel
    X_priv = _fillVariableSpace(
        X_informative,
        random_state,
        n_features=n_features,
        n_redundant=n_weakrel,
        n_strel=n_strel,
        n_repeated=n_repeated,
        partition=[n_weakrel],
    )

    if (
        problemName == "classification"
        or problemName == ProblemName.LUPI_CLASSIFICATION
    ):
        e = random_state.normal(
            size=(n_samples, X_priv.shape[1]), scale=0.65 / X_priv.std()
        )
        X = X_priv + e
        y = np.zeros_like(scores)
        y[scores > 0] = 1
        y[scores <= 0] = -1

    elif problemName == "regression" or problemName == ProblemName.LUPI_REGRESSION:
        e = random_state.normal(
            size=(n_samples, X_priv.shape[1]), scale=0.5 / X_priv.std()
        )
        X = X_priv + e
        y = scores
    elif (
        problemName == "ordinalRegression"
        or problemName == ProblemName.LUPI_ORDREGRESSION
    ):
        e = random_state.normal(
            size=(n_samples, X_priv.shape[1]), scale=0.2 / X_priv.std()
        )
        X = X_priv + e

        step = 1 / (n_ordinal_bins)
        quantiles = [i * step for i in range(1, n_ordinal_bins)]
        bs = np.quantile(scores, quantiles)
        bs = np.append(bs, np.inf)
        scores = scores[:, np.newaxis]
        y = np.sum(scores - bs >= 0, -1)

    if n_irrel > 0:
        X = np.hstack([X, random_state.normal(size=(n_samples, n_irrel))])
        X_priv = np.hstack([X_priv, random_state.normal(size=(n_samples, n_irrel))])

    if label_noise > 0:
        sample = random_state.choice(len(y), int(len(y) * label_noise))
        y[sample] = random_state.permutation(y[sample])

    return (X, X_priv, y.squeeze())



import matplotlib

matplotlib.use("TkAgg")
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from scipy.cluster.hierarchy import dendrogram
import matplotlib.cm as cm

# Get a color for each relevance type
color_palette_3 = cm.Set1([0, 1, 2], alpha=0.8)


def plot_relevance_bars(
    ax, ranges, ticklabels=None, classes=None, numbering=True, tick_rotation=30
):
    """

    Parameters
    ----------
    ax:
        axis which the bars get drawn on
    ranges:
        the 2d array of floating values determining the lower and upper bounds of the bars
    ticklabels: (optional)
        labels for each feature
    classes: (optional)
        relevance class for each feature, determines color
    numbering: bool
        Add feature index when using ticklabels
    tick_rotation:  int
        Amonut of rotation of ticklabels for easier readability.


    """
    N = len(ranges)

    # Ticklabels
    if ticklabels is None:
        ticks = np.arange(N) + 1
    else:
        ticks = list(ticklabels)
        if numbering:
            for i in range(N):
                ticks[i] += " - {}".format(i + 1)

    # Interval sizes
    ind = np.arange(N) + 1
    width = 0.6
    upper_vals = ranges[:, 1]
    lower_vals = ranges[:, 0]
    height = upper_vals - lower_vals
    # Minimal height to make very small intervals visible
    height[height < 0.001] = 0.001

    # Bar colors
    if classes is None:
        new_classes = np.zeros(N).astype(int)
        color = [color_palette_3[c.astype(int)] for c in new_classes]
    else:
        color = [color_palette_3[c.astype(int)] for c in classes]

    # Plot the bars
    bars = ax.bar(
        ind,
        height,
        width,
        bottom=lower_vals,
        tick_label=ticks,
        align="center",
        edgecolor=["black"] * N,
        linewidth=1.3,
        color=color,
    )

    ax.set_xticklabels(ticks)
    if ticklabels is not None:
        ax.set_xticklabels(ax.get_xticklabels(), rotation=tick_rotation, ha="right")
    # ax.tick_params(rotation="auto")
    # Limit the y range to 0,1 or 0,L1
    ax.set_ylim([0, max(ranges[:, 1]) * 1.1])

    ax.set_ylabel("relevance")
    ax.set_xlabel("feature")

    if classes is not None:
        relevance_classes = ["Irrelevant", "Weakly relevant", "Strongly relevant"]
        patches = []
        for i, rc in enumerate(relevance_classes):
            patch = mpatches.Patch(color=color_palette_3[i], label=rc)
            patches.append(patch)

        ax.legend(handles=patches)

    return bars


def plotIntervals(ranges, ticklabels=None, invert=False, classes=None):
    # Figure Parameters
    fig = plt.figure()
    ax = fig.add_subplot(111)

    out = plot_relevance_bars(ax, ranges, ticklabels=ticklabels, classes=classes)
    fig.autofmt_xdate()
    # Invert the xaxis for cases in which the comparison with other tools
    if invert:
        plt.gca().invert_xaxis()
    return fig


def plot_dendrogram_and_intervals(
    intervals, linkage, figsize=(13, 7), ticklabels=None, classes=None, **kwargs
):
    fig, (ax2, ax) = plt.subplots(2, 1, figsize=figsize)
    # Top dendrogram plot
    d = dendrogram(
        linkage,
        color_threshold=0,
        leaf_rotation=0.0,  # rotates the x axis labels
        leaf_font_size=12.0,  # font size for the x axis labels
        ax=ax2,
    )
    # Get index determined through linkage method and dendrogram
    rearranged_index = d["leaves"]
    ranges = intervals[rearranged_index]

    if ticklabels is None:
        ticks = np.array(rearranged_index)
        ticks += 1  # Index starting at 1
    else:
        ticks = list(ticklabels[rearranged_index])
        for i in range(len(intervals)):
            ticks[i] += " - {}".format(rearranged_index[i] + 1)

    plot_relevance_bars(
        ax,
        ranges,
        ticklabels=ticks,
        classes=classes[rearranged_index] if classes is not None else None,
        numbering=False,
        **kwargs,
    )
    fig.subplots_adjust(hspace=0)
    ax.margins(x=0)
    ax2.set_xticks([])
    ax2.margins(x=0)
    plt.tight_layout()

    return fig


def plot_intervals(model, ticklabels=None):
    """Plot the relevance intervals.
    
    Parameters
    ----------
    model : FRI model
        Needs to be fitted before.
    ticklabels : list of str, optional
        Strs for ticklabels on x-axis (features)
    """
    if model.interval_ is not None:
        plotIntervals(
            model.interval_, ticklabels=ticklabels, classes=model.relevance_classes_
        )
    else:
        print("Intervals not computed. Try running fit() function first.")


def plot_lupi_intervals(model, ticklabels=None, lupi_ticklabels=None):
    """Plot the relevance intervals.

        Parameters
        ----------
        model : FRI model
            Needs to be fitted before.
        ticklabels : list of str, optional
            Strs for ticklabels on x-axis (features)
        lupi_ticklabels : list of str, optional
            Strs for lupi ticklabels on x-axis (lupi features)
        """
    n_features = model.interval_.shape[0] - model.lupi_features_
    data_interval_ = model.interval_[0:n_features, :]
    lupi_interval_ = model.interval_[n_features:, :]

    data_relevance_classes_ = model.relevance_classes_[0:n_features]
    lupi_relevance_classes_ = model.relevance_classes_[n_features:]

    if model.interval_ is not None:
        plotIntervals(
            data_interval_, ticklabels=ticklabels, classes=data_relevance_classes_
        )
        plotIntervals(
            lupi_interval_, ticklabels=lupi_ticklabels, classes=lupi_relevance_classes_
        )
    else:
        print("Intervals not computed. Try running fit() function first.")


#
# def interactive_scatter_embed(embedding, mode="markers", txt=None):
#     # TODO: extend method
#     import plotly.graph_objs as go
#     from plotly.offline import init_notebook_mode, iplot
#     init_notebook_mode(connected=True)
#     # Create a trace
#     trace = go.Scatter(
#         x=embedding[:, 0],
#         y=embedding[:, 1],
#         mode=mode,
#         text=txt if mode is "text" else None
#     )
#
#     data = [trace]
#
#     # Plot and embed in ipython notebook!
#     iplot(data)



import numpy as np

from fri import ProblemName
from .gen_data import genRegressionData, genClassificationData, genOrdinalRegressionData
from .gen_lupi import genLupiData

__all__ = [
    "genRegressionData",
    "genClassificationData",
    "genOrdinalRegressionData",
    "genLupiData",
]


def quick_generate(problem: object, **kwargs) -> [np.ndarray, np.ndarray]:
    """
    Method to wrap individual data generation functions.
    Allows passing `problem` as a string such as "classification" or `ProblemName` object of the corresponding type.
    For possible kwargs see `genClassificationData' or `genLupiData`.

    Parameters
    ----------
    problem : str or `ProblemName`
        Type of data to generate (e.g. "classification" or `ProblemName.CLASSIFICATION`
    kwargs : **dict
        arguments to pass to the generation functions

    Returns
    -------
    Tuple[numpy.ndarray, numpy.ndarray]
    """
    if problem is "regression" or problem is ProblemName.REGRESSION:
        gen = genRegressionData
    elif problem is "classification" or problem is ProblemName.CLASSIFICATION:
        gen = genClassificationData
    elif problem is "ordreg" or problem is ProblemName.ORDINALREGRESSION:
        gen = genOrdinalRegressionData
    elif problem is "lupi_regression" or problem is ProblemName.LUPI_REGRESSION:
        gen = genLupiData
        kwargs["problemName"] = ProblemName.LUPI_REGRESSION
    elif problem is "lupi_classification" or problem is ProblemName.LUPI_CLASSIFICATION:
        gen = genLupiData
        kwargs["problemName"] = ProblemName.LUPI_CLASSIFICATION
    elif problem is "lupi_ordreg" or problem is ProblemName.LUPI_ORDREGRESSION:
        gen = genLupiData
        kwargs["problemName"] = ProblemName.LUPI_ORDREGRESSION
    else:
        raise ValueError(
            "Unknown problem type. Try 'regression', 'classification' or 'ordreg' and/or add 'lupi_' prefix"
        )
    return gen(**kwargs)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['fri', 'fri.model', 'fri.toydata']

package_data = \
{'': ['*'], 'fri': ['tests/*']}

install_requires = \
['cvxpy==1.0.24',
 'ecos>=2.0.5,<3.0.0',
 'joblib>=0.13.2,<0.14.0',
 'matplotlib>=3.1,<4.0',
 'numpy>=1.17,<2.0',
 'scikit-learn>=0.21.1,<0.22.0',
 'scipy>=1.0,<2.0']

setup_kwargs = {
    'name': 'fri',
    'version': '7.0.0rc0',
    'description': 'Implementation of Feature Relevance Bounds method to perform Feature Selection and further analysis.',
    'long_description': None,
    'author': 'Lukas Pfannschmidt',
    'author_email': 'lukas@lpfann.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://fri.lpfann.me',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import numpy as np
import pytest
from sklearn.preprocessing import scale
from sklearn.utils import check_random_state

from fri import FRI, ProblemName, quick_generate


@pytest.fixture(scope="module")
def randomstate():
    return check_random_state(1337)


@pytest.mark.parametrize(
    "problem",
    [ProblemName.REGRESSION, ProblemName.CLASSIFICATION, ProblemName.ORDINALREGRESSION],
)
def test__compute_single_preset_relevance_bounds(problem, randomstate):
    data = quick_generate(
        problem,
        n_samples=300,
        n_features=4,
        n_redundant=2,
        n_strel=2,
        random_state=randomstate,
    )

    X_orig, y = data
    X = scale(X_orig)

    model = FRI(problem, random_state=randomstate, n_jobs=1)
    model.fit(X, y)
    normal_range = model.interval_.copy()

    i = 0
    preset = [model.interval_[i, 0], model.interval_[i, 0]]
    range = model._relevance_bounds_computer.compute_single_preset_relevance_bounds(
        i, preset
    )
    assert normal_range.shape == range.shape
    assert range[i][0] == preset[0]
    assert range[i][1] == preset[1]


@pytest.mark.parametrize(
    "problem",
    [ProblemName.REGRESSION, ProblemName.CLASSIFICATION, ProblemName.ORDINALREGRESSION],
)
def test__compute_multi_preset_relevance_bounds(problem, randomstate):
    data = quick_generate(
        problem,
        n_samples=300,
        n_features=4,
        n_redundant=2,
        n_strel=2,
        random_state=randomstate,
    )

    X_orig, y = data
    X = scale(X_orig)

    model = FRI(problem, random_state=randomstate, n_jobs=1)
    model.fit(X, y)
    normal_range = model.interval_.copy()

    # We fix two feature values at the same time.
    # Note that we can not use the relevance bounds for multiple features, because of infeasibility
    # In the singular optimization case we exhaust the complete slack of our model constraints
    # If whe use multiple relevance bounds we double count this slack and the model optimization rightfully claims infeasible
    # Instead we use the mean value of the relevance bounds, which should be working in most cases
    i = 0
    mean_0 = np.mean([model.interval_[i, 0], model.interval_[i, 1]])
    preset_0 = [mean_0, mean_0]
    i = 1
    mean_1 = np.mean([model.interval_[i, 0], model.interval_[i, 1]])
    preset_1 = [mean_1, mean_1]
    presetModel = {0: preset_0, 1: preset_1}
    range = model._relevance_bounds_computer.compute_multi_preset_relevance_bounds(
        presetModel
    )

    assert normal_range.shape == range.shape
    i = 0
    assert range[i][0] == preset_0[0]
    assert range[i][1] == preset_0[1]
    i = 1
    assert range[i][0] == preset_1[0]
    assert range[i][1] == preset_1[1]


@pytest.mark.parametrize(
    "problem",
    [ProblemName.REGRESSION, ProblemName.CLASSIFICATION, ProblemName.ORDINALREGRESSION],
)
def test__compute_multi_preset_relevance_bounds(problem, randomstate):
    data = quick_generate(
        problem,
        n_samples=300,
        n_features=4,
        n_redundant=2,
        n_strel=2,
        random_state=randomstate,
    )

    X_orig, y = data
    X = scale(X_orig)

    model = FRI(problem, random_state=randomstate, n_jobs=1)
    model.fit(X, y)
    normal_range = model.interval_.copy()

    # We fix two feature values at the same time.
    # Note that we can not use the relevance bounds for multiple features, because of infeasibility
    # In the singular optimization case we exhaust the complete slack of our model constraints
    # If whe use multiple relevance bounds we double count this slack and the model optimization rightfully claims infeasible
    # Instead we use the mean value of the relevance bounds, which should be working in most cases
    i = 0
    mean_0 = np.mean([model.interval_[i, 0], model.interval_[i, 1]])
    preset_0 = [mean_0, mean_0]
    i = 1
    mean_1 = np.mean([model.interval_[i, 0], model.interval_[i, 1]])
    preset_1 = [mean_1, mean_1]
    presetModel = {0: preset_0, 1: preset_1}
    range = model.constrained_intervals(presetModel)

    assert normal_range.shape == range.shape
    i = 0
    assert range[i][0] == pytest.approx(preset_0[0])
    assert range[i][1] == pytest.approx(preset_0[1])
    i = 1
    assert range[i][0] == pytest.approx(preset_1[0])
    assert range[i][1] == pytest.approx(preset_1[1])


def test__compute_single_preset_relevance_bounds(randomstate):
    problem = ProblemName.CLASSIFICATION
    data = quick_generate(
        problem,
        n_samples=300,
        n_features=4,
        n_redundant=2,
        n_strel=2,
        random_state=randomstate,
    )

    X_orig, y = data
    X = scale(X_orig)

    model = FRI(problem, random_state=randomstate, n_jobs=1)
    model.fit(X, y)
    normal_range = model.interval_.copy()

    i = 0
    preset = [model.interval_[i, 0]]
    range = model._relevance_bounds_computer.compute_single_preset_relevance_bounds(
        i, preset
    )
    assert normal_range.shape == range.shape
    assert range[i][0] == preset[0]
    assert range[i][1] == preset

    preset = model.interval_[i, 0]
    range = model._relevance_bounds_computer.compute_single_preset_relevance_bounds(
        i, preset
    )
    assert normal_range.shape == range.shape
    assert range[i][0] == preset
    assert range[i][1] == preset



import numpy as np
import pytest
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_random_state

import fri
from fri.parameter_searcher import find_best_model
from fri.toydata import genLupiData


@pytest.fixture(scope="session")
def randomstate():
    return check_random_state(1337)


@pytest.mark.parametrize("n_weak", [0, 2])
@pytest.mark.parametrize("problem", fri.LUPI_MODELS)
def test_baseline_lupi(problem, n_weak, randomstate):
    n_samples = 300

    template = problem.value().get_initmodel_template
    params = problem.value().get_all_parameters()
    data = genLupiData(
        problem,
        n_strel=1,
        n_weakrel=n_weak,
        n_samples=n_samples,
        n_irrel=1,
        n_repeated=0,
        random_state=randomstate,
    )

    X, X_priv, y = data
    X = StandardScaler().fit(X).transform(X)
    X_priv = StandardScaler().fit(X_priv).transform(X_priv)
    combined = np.hstack([X, X_priv])

    iter = 50

    best_model, best_score = find_best_model(
        template,
        params,
        (combined, y),
        randomstate,
        iter,
        verbose=1,
        n_jobs=-2,
        lupi_features=X_priv.shape[1],
    )

    assert best_score > 0.5



import logging
from collections import defaultdict

import joblib
import numpy as np
from scipy import stats

from fri.model.base_cvxproblem import Relevance_CVXProblem
from fri.model.base_initmodel import InitModel
from fri.model.base_type import ProblemType
from fri.utils import permutate_feature_in_data

MIN_N_PROBE_FEATURES = 20  # Lower bound of probe features


def _start_solver_worker(bound: Relevance_CVXProblem):
    """
    Worker thread method for parallel computation
    """
    return bound.solve()


class RelevanceBoundsIntervals(object):
    def __init__(
        self,
        data,
        problem_type: ProblemType,
        best_init_model: InitModel,
        random_state,
        n_resampling,
        n_jobs,
        verbose,
        normalize=True,
    ):
        self.data = data
        self.problem_type = problem_type
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.n_resampling = n_resampling
        self.random_state = random_state
        self.best_init_model = best_init_model
        self.best_hyperparameters = best_init_model.hyperparam
        self.normalize = normalize

        # Relax constraints to improve stability
        relaxed_constraints = problem_type.get_relaxed_constraints(
            best_init_model.constraints
        )
        self.init_constraints = relaxed_constraints

    def get_normalized_lupi_intervals(self, lupi_features, presetModel=None):

        # We define a list of all the features we want to compute relevance bounds for
        X, _ = self.data  # TODO: handle other data formats
        all_d = X.shape[1]
        normal_d = all_d - lupi_features

        # Compute relevance bounds and probes for normal features and LUPI
        with joblib.Parallel(n_jobs=self.n_jobs, verbose=self.verbose) as parallel:
            d_n = _get_necessary_dimensions(normal_d, presetModel)
            rb = self.compute_relevance_bounds(d_n, parallel=parallel)
            probe_upper = self.compute_probe_values(d_n, True, parallel=parallel)
            probe_lower = self.compute_probe_values(d_n, False, parallel=parallel)

            d_l = _get_necessary_dimensions(all_d, presetModel, start=normal_d)
            rb_l = self.compute_relevance_bounds(d_l, parallel=parallel)
            probe_priv_upper = self.compute_probe_values(d_l, True, parallel=parallel)
            probe_priv_lower = self.compute_probe_values(d_l, False, parallel=parallel)
        probes = [probe_lower, probe_upper, probe_priv_lower, probe_priv_upper]
        #
        # Postprocess
        #

        # Get Scaling Parameters
        l1 = self.init_constraints["w_l1"]
        l1_priv = self.init_constraints["w_priv_l1"]
        l1 = l1 + l1_priv

        # Normalize Normal and Lupi features
        rb_norm = self._postprocessing(l1, rb)
        rb_l_norm = self._postprocessing(l1, rb_l)
        interval_ = np.concatenate([rb_norm, rb_l_norm])

        # Normalize Probes
        probe_lower = self._postprocessing(l1, probe_lower)
        probe_upper = self._postprocessing(l1, probe_upper)
        probe_priv_lower = self._postprocessing(l1, probe_priv_lower)
        probe_priv_upper = self._postprocessing(l1, probe_priv_upper)

        #
        #
        # Classify features
        fc = feature_classification(
            probe_lower, probe_upper, rb_norm, verbose=self.verbose
        )
        fc_l = feature_classification(
            probe_priv_lower, probe_priv_upper, rb_l_norm, verbose=self.verbose
        )
        fc_both = np.concatenate([fc, fc_l])

        return interval_, fc_both

    def get_normalized_intervals(self, presetModel=None):
        # We define a list of all the features we want to compute relevance bounds for
        X, _ = self.data  # TODO: handle other data formats
        d = X.shape[1]
        # Depending on the preset model, we dont need to compute all bounds
        # e.g. in the case of fixed features we skip those
        dims = _get_necessary_dimensions(d, presetModel)

        with joblib.Parallel(n_jobs=self.n_jobs, verbose=self.verbose) as parallel:
            relevance_bounds = self.compute_relevance_bounds(
                dims, parallel=parallel, presetModel=presetModel
            )
            probe_values_upper = self.compute_probe_values(
                dims, isUpper=True, parallel=parallel, presetModel=presetModel
            )
            probe_values_lower = self.compute_probe_values(
                dims, isUpper=False, parallel=parallel, presetModel=presetModel
            )

        # Postprocess bounds
        norm_bounds = self._postprocessing(
            self.best_init_model.L1_factor, relevance_bounds
        )
        norm_probe_values_upper = self._postprocessing(
            self.best_init_model.L1_factor, probe_values_upper
        )
        norm_probe_values_lower = self._postprocessing(
            self.best_init_model.L1_factor, probe_values_lower
        )
        feature_classes = feature_classification(
            norm_probe_values_lower,
            norm_probe_values_upper,
            norm_bounds,
            verbose=self.verbose,
        )

        return norm_bounds, feature_classes

    def compute_relevance_bounds(
        self, dims, parallel=None, presetModel=None, solverargs=None
    ):
        init_model_state = self.best_init_model.model_state

        work_queue = self._generate_relevance_bounds_tasks(
            dims, self.data, presetModel, init_model_state
        )

        # Solve relevance bounds in parallel (when available)
        if parallel is None:
            parallel = joblib.Parallel(n_jobs=self.n_jobs, verbose=self.verbose)
        bound_results = parallel(map(joblib.delayed(_start_solver_worker), work_queue))

        # Retrieve results and aggregate values in dict
        solved_bounds = defaultdict(list)
        for finished_bound in bound_results:

            # Only add bounds with feasible solutions
            if finished_bound.is_solved:
                solved_bounds[finished_bound.current_feature].append(finished_bound)

        # Initalize array for pair of bounds(= intervals)
        length = len(dims)
        intervals = np.zeros((length, 2))
        for abs_index, rel_index in zip(dims, range(length)):
            # Return interval for feature i (can be a fixed value when set beforehand)
            interval_i = self._create_interval(abs_index, solved_bounds, presetModel)
            intervals[rel_index] = interval_i

        return intervals  # TODO: add model model_state (omega, bias) to return value

    def compute_probe_values(self, dims, isUpper=True, parallel=None, presetModel=None):
        # Get model parameters
        init_model_state = self.best_init_model.model_state

        # Prepare parallel framework
        if parallel is None:
            parallel = joblib.Parallel(n_jobs=self.n_jobs, verbose=self.verbose)

        # Generate
        probe_queue = self._generate_probe_value_tasks(
            self.data,
            dims,
            isUpper,
            self.n_resampling,
            self.random_state,
            presetModel,
            init_model_state,
        )
        # Compute solution
        probe_results = parallel(map(joblib.delayed(_start_solver_worker), probe_queue))
        # probe_values.extend([probe.objective.value for probe in probe_results if probe.is_solved])

        candidates = defaultdict(list)
        for candidate in probe_results:
            # Only add bounds with feasible solutions
            if candidate.is_solved:
                candidates[candidate.probeID].append(candidate)
        probe_values = []
        for probes_for_ID in candidates.values():
            if isUpper:
                probe_values.append(
                    self.problem_type.get_cvxproblem_template.aggregate_max_candidates(
                        probes_for_ID
                    )
                )
            else:
                probe_values.append(
                    self.problem_type.get_cvxproblem_template.aggregate_min_candidates(
                        probes_for_ID
                    )
                )

        return np.array(probe_values)

    def _generate_relevance_bounds_tasks(
        self, dims, data, preset_model=None, best_model_state=None
    ):
        # Do not compute bounds for fixed features
        if preset_model is not None:
            dims = [di for di in dims if di not in preset_model]

        # Instantiate objects for computation later
        for di in dims:
            # Add Lower Bound problem(s) to work list
            yield from self.problem_type.get_cvxproblem_template.generate_lower_bound_problem(
                self.best_hyperparameters,
                self.init_constraints,
                best_model_state,
                data,
                di,
                preset_model,
            )

            # Add problem(s) for Upper bound
            yield from self.problem_type.get_cvxproblem_template.generate_upper_bound_problem(
                self.best_hyperparameters,
                self.init_constraints,
                best_model_state,
                data,
                di,
                preset_model,
            )

    def _generate_probe_value_tasks(
        self,
        data,
        dims,
        isUpper,
        n_resampling,
        random_state,
        preset_model=None,
        best_model_state=None,
    ):
        if isUpper:
            factory = (
                self.problem_type.get_cvxproblem_template.generate_upper_bound_problem
            )
        else:
            factory = (
                self.problem_type.get_cvxproblem_template.generate_lower_bound_problem
            )

        # Random sample n_resampling shadow features by permuting real features and computing upper bound
        random_choice = random_state.choice(a=dims, size=n_resampling)

        # Instantiate objects
        for i, di in enumerate(random_choice):
            data_perm = permutate_feature_in_data(data, di, random_state)

            # We only use upper bounds as probe features
            yield from factory(
                self.best_hyperparameters,
                self.init_constraints,
                best_model_state,
                data_perm,
                di,
                preset_model,
                probeID=i,
            )

    def _create_interval(
        self, feature: int, solved_bounds: dict, presetModel: dict = None
    ):
        # Return preset values for fixed features
        if presetModel is not None:
            if feature in presetModel:
                return presetModel[feature].squeeze()

        all_bounds = solved_bounds[feature]
        min_problems_candidates = [p for p in all_bounds if p.isLowerBound]
        max_problems_candidates = [p for p in all_bounds if not p.isLowerBound]
        if len(all_bounds) < 2:
            logging.error(
                f"(Some) relevance bounds for feature {feature} were not solved."
            )
            raise Exception("Infeasible bound(s).")
        lower_bound = self.problem_type.get_cvxproblem_template.aggregate_min_candidates(
            min_problems_candidates
        )
        upper_bound = self.problem_type.get_cvxproblem_template.aggregate_max_candidates(
            max_problems_candidates
        )
        return lower_bound, upper_bound

    def compute_single_preset_relevance_bounds(
        self, i: int, signed_preset_i: [float, float]
    ):
        """
        Method to run method once for one restricted feature
        Parameters
        ----------
        i:
            restricted feature
        signed_preset_i:
            restricted range of feature i (set before optimization = preset)

        """
        preset = {i: signed_preset_i}

        rangevector = self.compute_multi_preset_relevance_bounds(preset)

        return rangevector

    def compute_multi_preset_relevance_bounds(self, preset, lupi_features=0):
        """
        Method to run method with preset values

        Parameters
        ----------
        lupi_features
        """

        # The user is working with normalized values while we compute them unscaled
        if self.normalize:
            normalized = {}
            for k, v in preset.items():
                normalized[k] = np.asarray(v) * self.best_init_model.L1_factor
            preset = normalized

        # Add sign to presets
        preset = self._add_sign_to_preset(preset)

        # Calculate all bounds with feature i set to min_i
        if lupi_features > 0:
            rangevector, f_classes = self.get_normalized_lupi_intervals(
                lupi_features, presetModel=preset
            )
        else:
            rangevector, f_classes = self.get_normalized_intervals(presetModel=preset)

        return rangevector

    def _add_sign_to_preset(self, unsigned_presets):
        """
        We need signed presets for our convex problem definition later.
        We reuse the coefficients of the optimal model for this

        Parameters
        ----------
        unsigned_presets : dict

        Returns
        -------
        dict
        """

        signed_presets = {}
        # Obtain optimal model parameters
        w = self.best_init_model.model_state["w"]
        sum = 0
        for i, preset in unsigned_presets.items():
            preset = np.array(preset)
            if preset.size == 1:
                preset = np.repeat(preset, 2)
            unsigned_preset_i = np.sign(w[i]) * preset
            # accumulate maximal feature  contribution
            sum += unsigned_preset_i[1]  # Take upper preset
            signed_presets[i] = unsigned_preset_i

        # Check if unsigned_presets makes sense
        l1 = self.init_constraints["w_l1"]
        if sum > l1:
            print("maximum L1 norm of presets: ", sum)
            print("L1 allowed:", l1)
            print("Presets are not feasible. Try lowering values.")
            return

        return signed_presets

    def _postprocessing(self, L1, rangevector, round_to_zero=True):
        if self.normalize:
            assert L1 > 0
            rangevector = rangevector.copy() / L1

        if round_to_zero:
            rangevector[rangevector <= 1e-11] = 0
        return rangevector


def _get_necessary_dimensions(d: int, presetModel: dict = None, start=0):
    dims = np.arange(start, d)

    # if presetModel is not None:
    #    # Exclude fixed (preset) dimensions from being redundantly computed
    #    dims = [di for di in dims if di not in presetModel.keys()]
    # TODO: check the removal of this block
    return dims


def feature_classification(
    probes_low, probes_up, relevance_bounds, fpr=1e-4, verbose=0
):
    logging.debug("**** Feature Selection ****")
    logging.debug("Generating Lower Probe Statistic")
    lower_stat = create_probe_statistic(probes_low, fpr, verbose=verbose)
    logging.debug("Generating Upper Probe Statistic")
    upper_stat = create_probe_statistic(probes_up, fpr, verbose=verbose)

    weakly = relevance_bounds[:, 1] > upper_stat[1]
    strongly = relevance_bounds[:, 0] > lower_stat[1]
    both = np.logical_and(weakly, strongly)
    prediction = np.zeros(relevance_bounds.shape[0], dtype=np.int)
    prediction[weakly] = 1
    prediction[both] = 2

    return prediction


def create_probe_statistic(probe_values, fpr, verbose=0):
    # Create prediction interval statistics based on randomly permutated probe features (based on real features)
    n = len(probe_values)

    if n == 0:
        if verbose > 0:
            logging.info(
                "All probes were infeasible. All features considered relevant."
            )
        #    # If all probes were infeasible we expect an empty list
        #    # If they are infeasible it also means that only strongly relevant features were in the data
        #    # As such we just set the prediction without considering the statistics
        mean = 0
    else:
        probe_values = np.asarray(probe_values)
        mean = probe_values.mean()

    if mean == 0:
        lower_threshold, upper_threshold = mean, mean
        s = 0
    else:
        s = probe_values.std()
        lower_threshold = mean + stats.t(df=n - 1).ppf(fpr) * s * np.sqrt(1 + (1 / n))
        upper_threshold = mean - stats.t(df=n - 1).ppf(fpr) * s * np.sqrt(1 + (1 / n))

    if verbose > 0:
        print(
            f"FS threshold: {lower_threshold}-{upper_threshold}, Mean:{mean}, Std:{s}, n_probes {n}"
        )

    return lower_threshold, upper_threshold



from .classification import Classification
from .lupi_classification import LUPI_Classification
from .lupi_ordinal_regression import LUPI_OrdinalRegression
from .lupi_regression import LUPI_Regression
from .ordinal_regression import OrdinalRegression
from .regression import Regression

__all__ = [
    "Classification",
    "Regression",
    "OrdinalRegression",
    "LUPI_Classification",
    "LUPI_Regression",
    "LUPI_OrdinalRegression",
]



import numpy as np


def distance(u, v):
    """
    Distance measure custom made for feature comparison.

    Parameters
    ----------
    u: first feature
    v: second feature

    Returns
    -------

    """
    u = np.asarray(u)
    v = np.asarray(v)
    # Euclidean differences
    diff = (u - v) ** 2
    # Nullify pairwise contribution
    diff[u == 0] = 0
    diff[v == 0] = 0
    return np.sqrt(np.sum(diff))


def permutate_feature_in_data(data, feature_i, random_state):
    X, y = data
    X_copy = np.copy(X)
    # Permute selected feature
    permutated_feature = random_state.permutation(X_copy[:, feature_i])
    # Add permutation back to dataset
    X_copy[:, feature_i] = permutated_feature
    return X_copy, y



import cvxpy as cvx
import numpy as np
from sklearn import preprocessing
from sklearn.metrics import fbeta_score, classification_report
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import check_X_y
from sklearn.utils.multiclass import unique_labels

from fri.model.base_cvxproblem import Relevance_CVXProblem
from fri.model.base_initmodel import InitModel
from .base_type import ProblemType


class Classification(ProblemType):
    @classmethod
    def parameters(cls):
        return ["C"]

    @property
    def get_initmodel_template(cls):
        return Classification_SVM

    @property
    def get_cvxproblem_template(cls):
        return Classification_Relevance_Bound

    def relax_factors(cls):
        return ["loss_slack", "w_l1_slack"]

    def preprocessing(self, data, **kwargs):
        X, y = data
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit
        classes_ = unique_labels(y)

        if len(classes_) > 2:
            raise ValueError("Only binary class data supported")

        # Negative class is set to -1 for decision surface
        y = preprocessing.LabelEncoder().fit_transform(y)
        y[y == 0] = -1

        return X, y


class Classification_SVM(InitModel):
    @classmethod
    def hyperparameter(cls):
        return ["C"]

    def fit(self, X, y, **kwargs):
        (n, d) = X.shape

        C = self.hyperparam["C"]

        w = cvx.Variable(shape=(d), name="w")
        slack = cvx.Variable(shape=(n), name="slack")
        b = cvx.Variable(name="bias")

        objective = cvx.Minimize(cvx.norm(w, 1) + C * cvx.sum(slack))
        constraints = [cvx.multiply(y.T, X * w + b) >= 1 - slack, slack >= 0]

        # Solve problem.
        solver_params = self.solver_params
        problem = cvx.Problem(objective, constraints)
        problem.solve(**solver_params)

        w = w.value
        b = b.value
        slack = np.asarray(slack.value).flatten()
        self.model_state = {"w": w, "b": b, "slack": slack}

        loss = np.sum(slack)
        w_l1 = np.linalg.norm(w, ord=1)
        self.constraints = {"loss": loss, "w_l1": w_l1}
        return self

    def predict(self, X):
        w = self.model_state["w"]
        b = self.model_state["b"]
        y = np.dot(X, w) + b >= 0
        y = y.astype(int)
        y[y == 0] = -1
        return y

    def score(self, X, y, **kwargs):
        prediction = self.predict(X)

        # Negative class is set to -1 for decision surface
        y = LabelEncoder().fit_transform(y)
        y[y == 0] = -1

        # Using weighted f1 score to have a stable score for imbalanced datasets
        score = fbeta_score(y, prediction, beta=1, average="weighted")
        if "verbose" in kwargs:
            return classification_report(y, prediction)
        return score


class Classification_Relevance_Bound(Relevance_CVXProblem):
    def init_objective_UB(self, sign=None, **kwargs):
        self.add_constraint(
            self.feature_relevance <= sign * self.w[self.current_feature]
        )
        self._objective = cvx.Maximize(self.feature_relevance)

    def init_objective_LB(self, **kwargs):
        self.add_constraint(
            cvx.abs(self.w[self.current_feature]) <= self.feature_relevance
        )
        self._objective = cvx.Minimize(self.feature_relevance)

    def _init_constraints(self, parameters, init_model_constraints):
        # Upper constraints from initial model
        l1_w = init_model_constraints["w_l1"]
        init_loss = init_model_constraints["loss"]

        C = parameters["C"]

        # New Variables
        self.w = cvx.Variable(shape=(self.d), name="w")
        self.b = cvx.Variable(name="b")
        self.slack = cvx.Variable(shape=(self.n), nonneg=True, name="slack")

        # New Constraints
        distance_from_plane = cvx.multiply(self.y, self.X * self.w + self.b)
        self.loss = cvx.sum(self.slack)
        self.weight_norm = cvx.norm(self.w, 1)

        self.add_constraint(distance_from_plane >= 1 - self.slack)
        self.add_constraint(self.weight_norm <= l1_w)
        self.add_constraint(C * self.loss <= C * init_loss)

        self.feature_relevance = cvx.Variable(nonneg=True, name="Feature Relevance")
